package xic.analysis;

import java.util.HashSet;
import java.util.Set;

import xic.codegen.RegAlloc;
import xic.ir.IRCall;
import xic.ir.IRMem;
import xic.ir.IRMove;
import xic.ir.IRNode;
import xic.ir.IRTemp;


/**
 * IR-level analysis for the detection of essential variables, i.e. 
 * variables which have an impact on the behavior of a function.
 */
public class EssentialVariables extends DataFlowAnalysis {
	public static class EVValue implements Value {
		/* List of variable names */
		private Set<String> vars;

		public EVValue() {
			vars = new HashSet<String>();
		}

		public EVValue(Set<String> vars) {
			this.vars = new HashSet<String>(vars);
		}

		public Value copy() {
			return new EVValue(vars);
		}

		public void meet(Value v) {
			vars.addAll(((EVValue) v).vars);
		}
		
		public String toString() {
			return vars.toString();
		}

		public int hashCode() {
			return vars.hashCode();
		}
		
		public Set<String> vars() {
			return vars;
		}
		
		public boolean equals(Object obj) {
			if (!(obj instanceof EVValue))
				return false;
			return vars.equals(((EVValue) obj).vars);
		}

		public void setNullMeet() {
			vars.clear();
			vars.add(RegAlloc.RV_NAME);
		}
	};

	/// Dataflow-related information associated with each node
	public static class EVFlowInfo extends FlowInfo {
		public Set<String> def, use;
		public IRNode node;
		public boolean isCallAbort;

		public EVFlowInfo(Object o) {
			use = new HashSet<String>();
			def = new HashSet<String>();
			entry = new EVValue();
			exit = new EVValue();
			
			if (o instanceof IRNode) {
				node = (IRNode) o;
				node.findWrittenTemps(def);
				node.findReadTemps(use);	
			}

			isCallAbort = o.equals("call_abort");
		}
	};
	
	public EssentialVariables(Set<CFGNode> nodes) {
		super(nodes);
		
		dir = Direction.BACKWARD;
		top = new EVValue();
	}

	protected FlowInfo createFlowInfo(CFGNode node) {
		return new EVFlowInfo(node.value());
	}

	public Set<String> getEssentialVariablesBefore(CFGNode node) {
		return ((EVValue) ((EVFlowInfo) getFlowInfo(node)).exit).vars;
	}

	public Set<String> getEssentialVariablesAfter(CFGNode node) {
		return ((EVValue) ((EVFlowInfo) getFlowInfo(node)).entry).vars;
	}

	public void F(FlowInfo flowInfo) {
		EVFlowInfo f = (EVFlowInfo) flowInfo;
		Set<String> exit = ((EVValue) f.exit).vars;
		Set<String> entry = ((EVValue) f.entry).vars;
		boolean makeUseEssential = false;

		if (f.isCallAbort)
			entry.clear(); // call_abort doesn't depend on RV

		exit.clear();
		exit.addAll(entry);
		exit.removeAll(f.def);

		if (f.node != null && f.node instanceof IRMove) {
			IRMove move = (IRMove) f.node;
			
			if (move.target() instanceof IRMem) {
				makeUseEssential = true;
			} else {
				if (entry.contains(((IRTemp) move.target()).name())) {
					makeUseEssential = true;
				} else if (move.expr() instanceof IRCall) {
					makeUseEssential = true;
				}
			}
		} else {
			exit.addAll(f.use);
		}
		
		if (makeUseEssential)
			exit.addAll(f.use);
	}
}