package xic.analysis;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.PriorityQueue;
import java.util.Set;

import xic.Configuration;
import xic.InternalCompilerException;
import xic.analysis.WebGenerator.Web;


/**
 * Some utilities for performing optimistic coloring of web interference graphs
 * and generating visualizations thereof.
 */
public class GraphColoringUtils {
	public static final boolean DEBUG = false;
	public static int UNINITIALIZED_COLOR = -1;
	public static int IGNORE_COLOR = -2;

	public static boolean findOptimisticColoring(
			int nColors, int availableColors, 
			PriorityQueue<Web> pq) {
		Web web = pq.poll();

		if (web == null)
			return true; // Trivial

		web.setColor(IGNORE_COLOR); // We'll figure out the color later
		for (Web w : web.neighbors()) {
			// Degree has changed
			if (pq.contains(w)) {
				pq.remove(w);
				pq.add(w);
			}
		}

		/* Coloring may have failed in a recursive call - check if this is the case */
		if (!findOptimisticColoring(nColors, availableColors, pq))
			return false;

		if (web.color() >= 0)
			throw new InternalCompilerException("Internal error while coloring the interference graph");

		/* Preferred colors due to constraints of neighbors */
		int preferredColors = availableColors;

		/* Preferred colors due to own bias */
		int preferredColorsSelf = availableColors;

		if (web.hasRegisterPreferences()) {
			preferredColorsSelf = 0;
			for (String regName : web.registerPreferences())
				preferredColorsSelf |= 1 << Configuration.registerIndex(regName);
		}

		for (Web n : web.neighbors()) {
			if (!n.isNeighborOf(web))
				throw new InternalCompilerException("Graph structure inconsistency");

			int neighborColor = n.color();
			if (neighborColor == -1) {
				throw new InternalCompilerException(
						"Internal error while coloring the interference graph!");
			} else if (neighborColor == IGNORE_COLOR) { // "Removed" from the graph
				if (n.hasRegisterPreferences()) {
					/* Try to honor n's preference */
					for (String regName : n.registerPreferences())
						preferredColors &= ~(1 << Configuration.registerIndex(regName));
				}
				continue;
			}

			availableColors &= ~(1 << neighborColor);
		}

		if (DEBUG) {
			System.out.println("Coloring web " + web);
			System.out.println("Available colors             : " + bitVectorToString(availableColors));
			System.out.println("Preferred colors (neighbors) : " + bitVectorToString(preferredColors));
			System.out.println("Preferred colors (self)      : " + bitVectorToString(preferredColorsSelf));
		}

		/* First, try to find a color that makes everyone happy */
		for (int i=0; i<nColors; ++i) {
			if ((availableColors & (1 << i)) == 0
				|| (preferredColors & (1 << i)) == 0
				|| (preferredColorsSelf & (1 << i)) == 0)
				continue;
			web.setColor(i);
			return true;
		}

		/* If that didn't work, try to find a color that only makes the current node happy */
		for (int i=0; i<nColors; ++i) {
			if ((availableColors & (1 << i)) == 0
				|| (preferredColorsSelf & (1 << i)) == 0)
				continue;
			web.setColor(i);
			return true;
		}

		/* If that didn't work, try to find a color that only makes the neighbors happy */
		for (int i=0; i<nColors; ++i) {
			if ((availableColors & (1 << i)) == 0
				|| (preferredColorsSelf & (1 << i)) == 0)
				continue;
			web.setColor(i);
			return true;
		}

		/* Stop being picky -- any color will do */
		for (int i=0; i<nColors; ++i) {
			if ((availableColors & (1 << i)) == 0)
				continue;
			web.setColor(i);
			return true;
		}

		if (web.degree() < nColors)
			throw new InternalCompilerException(
					"Internal error while coloring the interference graph (a bit twiddling problem, perhaps?)!");

		/* We have a spill despite optimism :( 
		   - consider the offender and its neighbors as candidates for spilling
		 */
		pq.clear();
		pq.add(web);
		for (Web n : web.neighbors()) {
			if (n.defs().size() > 0 && !Configuration.REG_NAMES.contains(n.varNameString()))
				pq.add(n);
		}
		return false;
	}
	
	private static String bitVectorToString(int registers) {
		StringBuffer result = new StringBuffer("{");
		boolean first = true;
		for (int i=0; i<Configuration.AVAILABLE_REGISTERS.length; ++i) {
			if ((registers & (1 << i)) == 0)
				continue;
			if (!first)
				result.append(", ");
			result.append(Configuration.AVAILABLE_REGISTERS[i].toString());
			first = false;
		}
		return result.append("}").toString();
	}

	public static String dumpDot(Collection<Web> nodes) {
		StringBuffer buf = new StringBuffer();

		String colors[] = new String[] {
			"red", "black",
			"white", "black",
			"darkslateblue", "white",
			"forestgreen", "black",
			"indianred", "black",
			"black", "white",
			"beige", "black",
			"gold", "black",
			"orangered", "black",

			"red", "gray",
			"white", "gray",
			"darkslateblue", "gray",
			"forestgreen", "gray",
			"indianred", "gray",
			"black", "gray",
			"beige", "gray",
			"gold", "gray",
			"gray", "black",
			"blue", "white"
		};

		buf.append("graph dot_representation {\n" +
				"\trankdir = TB;\n" +
				"\tsize = \"8,5\"\n" +
				"\tnode [shape = rectangle]\n");
		Set<String> edges = new HashSet<String>();
		Map<Web, String> nameMap = new HashMap<Web, String>();
		int counter = 0;

		for (Web node : nodes) {
			String nodeName = null;

			if (nameMap.containsKey(node)) {
				nodeName = nameMap.get(node);
			} else {
				nodeName = "N" + counter++;
				nameMap.put(node, nodeName);
			}
			String label = node.toString().trim().replaceAll("\\n", "\\\\l").
				replaceAll("\\r", "").replaceAll("\"", "\\\\\"") + "\\l";
			int color = node.color();

			if (node.isSpilled())
				color = -2;
			else if (node.color() == GraphColoringUtils.IGNORE_COLOR)
				color = -1;


			buf.append("\t" + nodeName + "[style=filled "+
					"label=\""+label+"\" " +
					"fillcolor="+colors[(color+2)*2]+" " +
					"fontcolor="+colors[(color+2)*2+1]+"]"+ ";\n");
			for (Web neighbor: node.neighbors()) {
				String neighborName = null;
				if (nameMap.containsKey(neighbor)) {
					neighborName = nameMap.get(neighbor);
				} else {
					neighborName = "N" + counter++;
					nameMap.put(neighbor, neighborName);
				}
				if (edges.contains( "\t" + neighborName + " -- " + nodeName + ";"))
					continue; // no double edges
				String edgeName = "\t" + nodeName + " -- " + neighborName + ";";
				edges.add(edgeName);
				buf.append(edgeName).append('\n');
			}
		}
		buf.append("}\n");
		return buf.toString();
	}
	
}
