package xic.analysis;

import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import xic.InternalCompilerException;
import xic.analysis.WebGenerator.Web;
import xic.ir.NameGenerator;


/**
 * Utility functions for decomposing a control flow graph
 * into a tree of nested loops
 */
public class LoopStructure {
	public static class Loop {
		private Set<CFGNode> nodes;
		private List<Loop> children;
		private CFGNode head;
		private boolean isLoop;

		public Loop(CFGNode head, Set<CFGNode> nodes) {
			this.nodes = nodes;
			children = new ArrayList<Loop>();
			this.head = head;
			isLoop = true;
		}

		public Loop() {
			children = new ArrayList<Loop>();
			isLoop = false;
		}

		public Set<CFGNode> getNodes() {
			return nodes;
		}
		
		public List<Loop> getChildren() {
			return children;
		}
		
		public boolean isLoop() {
			return isLoop;
		}

		public void addChild(Loop loop) {
			children.add(loop);
		}

		public String toDot() {
			StringBuffer dotSource = new StringBuffer(
					"digraph dot_representation {\n" +
					"rankdir = TB;\n" +
					"size = \"8,5\"\n" +
					"root [shape = circle];\n" +
					"node [shape = rectangle];\n"
			);
			NameGenerator nameGen = new NameGenerator("N", "cluster");

			for (Loop l : children)
				l.toDot(dotSource, "root", nameGen);
			dotSource.append("}");
			return dotSource.toString();
		}

		private void toDot(StringBuffer dotSource, String rootName, NameGenerator nameGen) {
			String subgraphName = nameGen.nextTemp();
			dotSource.append("subgraph " + subgraphName + "{\n");
			dotSource.append("style=filled;\n");
			dotSource.append("color=lightgrey;\n");

			DotGenerator dg = new DotGenerator(nameGen);
			for (CFGNode n : nodes)
				dotSource.append(dg.generateDotNode(n));
			dotSource.append("}\n");
			String headName = dg.getName(head);
			dotSource.append(rootName + " -> " + headName + ";\n");
			for (Loop l2 : children)
				l2.toDot(dotSource, headName, nameGen);
		}

		public CFGNode getHead() {
			return head;
		}

		public void organize(Dominators d) {
			List<Loop> innerLoops = new ArrayList<Loop>();
			boolean restart;
			do {
				/* This is terrible complexity-wise, but fortunately 
				    there should not be too many loops .. */
				restart = false;
				for (Loop outerLoop : children) {
					innerLoops.clear();

					boolean isOuterLoop = true;
					for (Loop innerLoop : children) {
						if (innerLoop == outerLoop)
							continue;
						if (innerLoop.getNodes().contains(outerLoop.getHead())) {
							isOuterLoop = false;
							break;
						} else if (outerLoop.getNodes().contains(innerLoop.getHead())) {
							innerLoops.add(innerLoop);
						}
					}

					if (isOuterLoop && innerLoops.size() > 0) {
						children.removeAll(innerLoops);
						outerLoop.children.addAll(innerLoops);
						outerLoop.organize(d);
						restart = true;
						break;
					}
					
					if (restart)
						break;
				}
			} while (restart);
		}
		
		public String toString() {
			return String.valueOf(head);
		}
	}

	private Map<Object, Integer> loopDepth;
	private Loop root;

	public LoopStructure(CFGNode entry) {
		loopDepth = new HashMap<Object, Integer>();
		root = getLoopStructure(entry);
		computeLoopDepth(root, 0);
	}

	/**
	 * Look up the loop depth for the given CFG node
	 */
	public int getLoopDepth(Object node) {
		if (!loopDepth.containsKey(node))
			return 0;
		return loopDepth.get(node);
	}
	
	/**
	 * Annotate CFG nodes referenced by webs with
	 * loop depth information
	 */
	public void annotate(Collection<Web> webs) {
		for (Web w : webs) {
			for (CFGNode n : w.uses())
				w.setLoopDepth(n, getLoopDepth(n.value()));
			for (CFGNode n : w.defs())
				w.setLoopDepth(n, getLoopDepth(n.value()));
		}
	}

	private void computeLoopDepth(Loop loop, int depth) {
		for (Loop childLoop : loop.getChildren())
			computeLoopDepth(childLoop, depth+1);

		if (loop.isLoop()) {
			for (CFGNode n : loop.getNodes()) {
				if (loopDepth.containsKey(n.value())) {
					int previousDepth = loopDepth.get(n.value());
					if (previousDepth >= depth)
						continue;
				}
				loopDepth.put(n.value(), depth);
			}
		}
	}


	/**
	 * Generate GraphViz files for each function in a program
	 */
	public static void dumpLoops(Map<String, CFGNode> finalCFGs) {
		try {
			for (String funcName : finalCFGs.keySet()) {
				CFGNode rootNode = finalCFGs.get(funcName);
				PrintWriter pw = new PrintWriter(new FileWriter("loops_"
						+ funcName + ".dot"));
				pw.print(getLoopStructure(rootNode).toDot());
				pw.close();
			}
		} catch (IOException e) {
			throw new InternalCompilerException(
					"Could not generate debug output", e);
		}
	}
 
	/**
	 * Compute the loop structure starting from the entry point of a function
	 */
	public static Loop getLoopStructure(CFGNode entry) {
		Dominators dominators = new Dominators(entry.flatten());
		dominators.solve();

		Loop rootNode = new Loop();

		findLoops(dominators, new HashSet<CFGNode>(), entry, 
				rootNode.getChildren());
		rootNode.organize(dominators);

		return rootNode;
	}

	/**
	 * Search for back-edges and take note of loops
	 */
	private static void findLoops(Dominators d, Set<CFGNode> seen, CFGNode node, List<Loop> loops) {
		seen.add(node);
		for (CFGNode n : node.outEdges()) {
			if (!d.getDominators(n).contains(node)) {
				if (n.value() instanceof String) // ignore special nodes (start, end, call_abort)
					continue;
				Set<CFGNode> nodes = new HashSet<CFGNode>();
				exploreLoop(d, n, n, nodes);
				loops.add(new Loop(n, nodes));
				continue;
			} else {
				findLoops(d, seen, n, loops);
			}
		}
	}

	/**
	 * Explore a loop in the reverse graph
	 */
	private static void exploreLoop(Dominators d, CFGNode head, CFGNode node,
			Set<CFGNode> loopNodes) {
			
		if (loopNodes.contains(node))
			return;
		loopNodes.add(node);
		for (CFGNode n : node.inEdges()) {
			if (n == head)
				continue;
			if (!d.getDominators(n).contains(head))
				continue;
			exploreLoop(d, head, n, loopNodes);
		}
	}
}
