package xic.ast;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.StringReader;
import java.util.Iterator;

import xic.ErrorSink;

import edu.cornell.cs.cs4120.xi.CompilationException;
import edu.cornell.cs.cs4120.xi.Position;
import edu.cornell.cs.cs4120.util.VisualizableTreeNode;

/**
 * Contains various utility functions used to run test
 * cases and provide error reporting.
 */
public class ErrorReporter {
	/**
	 * This function checks two ASTs against each other. Optionally, posotion
	 * records can be included in the comparison.
	 */
	public static void checkNodes(VisualizableTreeNode n1,
			VisualizableTreeNode n2, boolean checkPos) {
		assertNotNull(n1.children());
		assertNotNull(n2.children());
		Iterator<VisualizableTreeNode> n1_children = n1.children().iterator();
		Iterator<VisualizableTreeNode> n2_children = n2.children().iterator();

		assertEquals("Node label mismatch in AST", n1.label(), n2.label());
		assertEquals("Child count mismatch in AST node '" + n1.label() + "':",
				n1_children.hasNext(), n2_children.hasNext());

		if (checkPos)
			assertEquals("Position mismatch in AST node '" + n1.label() + "':",
					((ASTNode) n1).position(), ((ASTNode) n2).position());

		if (n1_children.hasNext()) {
			while (n1_children.hasNext() && n2_children.hasNext()) {
				ASTNode n1_child = (ASTNode) n1_children.next();
				ASTNode n2_child = (ASTNode) n2_children.next();

				checkNodes(n1_child, n2_child, checkPos);
			}
			assertEquals("Child count mismatch in AST node '" + n1.label()
					+ "':", n1_children.hasNext(), n2_children.hasNext());
		} 
	}

	/**
	 * Takes a line of source code and a position (measured in characters
	 * from the left side). The output is a string which, when printed
	 * next to the source code line, will align the given character with
	 * the specified position. This function mainly exists to handle issues
	 * with tab characters. 
	 */
	public static String makeHighlightString(String string, int pos, char c) {
		StringBuffer buffer = new StringBuffer();
		for (int i=0; i<pos; ++i) {
			if (string.charAt(i) == '\t')
				buffer.append('\t');
			else
				buffer.append(' ');
		}
		buffer.append(c);
		return buffer.toString();
	}

	public static void printErrors(ErrorSink sink) throws IOException {
		for (CompilationException e: sink.getErrors())
			errPrettyPrint(e.getPosition(), e.getMessage(), null);
	}
	
	public static void printErrors(String source, ErrorSink sink) throws IOException {
		for (CompilationException e: sink.getErrors())
			errPrettyPrint(e.getPosition(), e.getMessage(), source);
	}


	public static void errPrettyPrint(Position p, String message, String source) throws IOException {
		boolean isSyntaxError = message.equals("syntax error") || message.startsWith("Unable to recover");
		BufferedReader reader;
	
		if (p == null) {
			System.out.println("Error without position information: " + message);
			return;
		} else if (source != null) {
			reader = new BufferedReader(new StringReader(source));
		} else {
			FileInputStream is = new FileInputStream(new File(p.unit()));
			InputStreamReader isr = new InputStreamReader(is);
			reader = new BufferedReader(isr);
		}
		errPrettyPrint(reader, p, isSyntaxError ? null : message, !isSyntaxError);
	}

	public static void errPrettyPrint(BufferedReader input, Position p, String message, boolean semantic) throws IOException {
		/* Skip preceding lines */
		for(int i = 0; i < p.lineStart(); i++)
			input.readLine();

		boolean sameLine = (p.lineStart() == p.lineEnd());	
		boolean sameColumn = (p.columnStart() == p.columnEnd());
		String errorType = semantic ? "semantic" : "syntax";

		if (message == null)
			message = "";
		else
			message = "\"" + message + "\"";

		if (p.columnEnd() == -1) {
			System.out.println(p.unit() + ": fatal "+errorType+" error at line "
					+ p.lineStart() + ": " + message);
			return;
		}

		if (sameLine && sameColumn) {
			System.out.println(p.unit() + ": "+ errorType +" error at position "
					+ p.toString() + ": " + message);
			String line = input.readLine();
			System.out.println("\t"+line);
			System.out.println("\t"+makeHighlightString(line, p.columnStart(), '^'));
		} else {
			
			System.out.println(p.unit()
					+ ": "+errorType+" error between positions " + p.toString()
					+ ": " + message);

			String line = input.readLine();
			System.out.println("\t"+makeHighlightString(line, p.columnStart(), 'v'));
			System.out.println("\t"+line);

			if (!sameLine) {
				for (int i=p.lineStart(); i<p.lineEnd()-1; ++i)
						System.out.println("\t" + input.readLine());		
				line = input.readLine();
				System.out.println("\t"+line);
			}
			System.out.println("\t"+makeHighlightString(line, p.columnEnd(), '^'));
		}

		input.close();
	}
}
