package xic.codegen;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import xic.InternalCompilerException;
import xic.codegen.matchers.BinOpMatcher;
import xic.codegen.matchers.CJumpMatcher;
import xic.codegen.matchers.CallMatcher;
import xic.codegen.matchers.CompUnitMatcher;
import xic.codegen.matchers.ConstMatcher;
import xic.codegen.matchers.CtorMatcher;
import xic.codegen.matchers.DataSeqMatcher;
import xic.codegen.matchers.DeclMatcher;
import xic.codegen.matchers.FuncDeclMatcher;
import xic.codegen.matchers.IncDecMatcher;
import xic.codegen.matchers.JumpMatcher;
import xic.codegen.matchers.LabelMatcher;
import xic.codegen.matchers.LeaMatcher;
import xic.codegen.matchers.MemMatcher;
import xic.codegen.matchers.MoveMatcher;
import xic.codegen.matchers.NameMatcher;
import xic.codegen.matchers.SeqMatcher;
import xic.codegen.matchers.TempMatcher;
import xic.codegen.matchers.UnaryOpMatcher;
import xic.codegen.matchers.ZDataMatcher;
import xic.ir.IRNode;
import xic.ir.IRPrinter;


/**
 * The tiler class oversees the translation of
 * an IR tree to a set of optimal tiles. It is
 * responsible for the memoization of tile 
 * candidates.
 */
public class Tiler {
	private Map<IRNode, Tile> cache;
	private List<Matcher> matchers;
	private Set<String> boostedMatchers;
	private boolean debug = false;
	private int debugCtr = 0;
	
	public Tiler() {
		this(new HashSet<String>());
	}

	public Tiler(Set<String> boostedMatchers) {
		this.boostedMatchers = boostedMatchers;
		cache = new HashMap<IRNode, Tile>();
		matchers = new ArrayList<Matcher>();

		/* Register all implemented matchers */
		matchers.add(new IncDecMatcher());
		matchers.add(new UnaryOpMatcher());
		matchers.add(new MoveMatcher());
		matchers.add(new SeqMatcher());
		matchers.add(new FuncDeclMatcher());
		matchers.add(new CompUnitMatcher());
		matchers.add(new TempMatcher());
		matchers.add(new LabelMatcher());
		matchers.add(new ConstMatcher());
		matchers.add(new MemMatcher());
		matchers.add(new JumpMatcher());
		matchers.add(new BinOpMatcher());
		matchers.add(new LeaMatcher());
		matchers.add(new CJumpMatcher());
		matchers.add(new CallMatcher());
		matchers.add(new NameMatcher());
		matchers.add(new DataSeqMatcher());
		matchers.add(new ZDataMatcher());
		matchers.add(new CtorMatcher());
		matchers.add(new DeclMatcher());
	}

	public Tile tile(IRNode node) {
		if (cache.containsKey(node))
			return cache.get(node);
		
		int ctr = debugCtr;

		if (debug)
			System.out.println(">>>>> Tiling " + IRPrinter.IRToString(node));

		int bestCost = Integer.MAX_VALUE;
		Tile bestTile = null;
		for (Matcher m : matchers) {
			Tile tile = m.match(node, this);
			if (tile == null)
				continue;
			
			int estimatedCost = tile.estimateCost();

			if (boostedMatchers.contains(m.getClass().getSimpleName()))
				estimatedCost = 0;

			if (estimatedCost < bestCost) {
				bestTile = tile;
				bestCost = estimatedCost;
			}

			if (debug) {
				if (ctr != debugCtr)
					System.out.println("===== Continuing to tile " + IRPrinter.IRToString(node));
				ctr = ++debugCtr;
			}

			if (debug)
				System.out.println(" * " + m.getClass().getSimpleName()
						+ " proposes a tile of cost " + estimatedCost);
		}

		if (bestTile == null)
			throw new InternalCompilerException(
				"Unable to tile " + IRPrinter.IRToString(node));
		
		if (debug)
			System.out.println("<<<<< Done tiling " + IRPrinter.IRToString(node));

		cache.put(node, bestTile);
		
		return bestTile;
	}
}
