package xic.codegen.matchers;

import java.util.ArrayList;
import java.util.List;

import xic.Configuration;
import xic.codegen.AbstractTile;
import xic.codegen.Instr;
import xic.codegen.InstrList;
import xic.codegen.Matcher;
import xic.codegen.RegAlloc;
import xic.codegen.Tile;
import xic.codegen.Tiler;
import xic.codegen.matchers.AddrModeHelper.AddrMode;
import xic.ir.IRCall;
import xic.ir.IRExpr;
import xic.ir.IRMem;
import xic.ir.IRName;
import xic.ir.IRNode;
import xic.ir.NameGenerator;


public class CallMatcher implements Matcher {
	private static class CallTile extends AbstractTile {
		private String callTarget = null;
		private AddrMode addrMode = null;
		private Tile callTargetTile = null;
		private boolean isProcedure;

		public CallTile(String callTarget, boolean isProcedure, List<Tile> paramTiles) {
			this.callTarget = callTarget;
			this.isProcedure = isProcedure;
			children.addAll(paramTiles);
		}

		public CallTile(Tile target, boolean isProcedure, List<Tile> paramTiles) {
			this.callTargetTile = target;
			this.isProcedure = isProcedure;
			children.addAll(paramTiles);		
		}

		public CallTile(AddrMode addrMode, boolean isProcedure, List<Tile> paramTiles) {
			this.addrMode = addrMode;
			this.isProcedure = isProcedure;
			children.addAll(paramTiles);
		}

		public int localCost() {
			int cost = 1;
			
			if (callTargetTile != null)
				cost += callTargetTile.estimateCost();
			
			if (children.size() > 0)
				cost += children.size() + 1;

			if(addrMode != null)
				cost += addrMode.estimateCost();

			return cost;
		}

		public String translate(InstrList il, NameGenerator ng) {
			int argIdx = children.size(), regCtr = 0, stackCtr = 0;
			int nStackArgs = Math.max(0,
					children.size() - Configuration.PARAMETER_REGISTERS.length);

			for (Tile tile : children) {
				Instr.Op source = null;
				boolean memOp = false;
				if (tile instanceof ConstMatcher.ConstTile) {
					source = new Instr.Op(((ConstMatcher.ConstTile) tile).value());
				} else if (tile instanceof MemMatcher.MemTile) {
					source = ((MemMatcher.MemTile) tile).getOp(il, ng);
					memOp = true;
				} else {
					String tempName = tile.translate(il, ng);
					source = new Instr.Op(Instr.Op.Type.REG, tempName);
				}

				if (--argIdx < Configuration.PARAMETER_REGISTERS.length) {
					Instr.Op dest = new Instr.Op(Instr.Op.Type.REG, 
							Configuration.PARAMETER_REGISTERS[argIdx]);
					il.add(new Instr("mov", dest, source));
					regCtr++;
				} else {
					if (memOp) {
						String tempName = ng.nextTemp();
						il.add(new Instr("mov", new Instr.Op(Instr.Op.Type.REG, tempName), source));
						source = new Instr.Op(Instr.Op.Type.REG, tempName);
					}
					stackCtr++;
					Instr.Op dest = new Instr.Op(RegAlloc.CALL_ARG_SCRATCH, null, 0, 
							(nStackArgs-stackCtr) * Configuration.WORD_SIZE);
					il.add(new Instr("mov", dest, source));
				}
			}


			Instr callInstr = null;
			if (callTarget != null) {
				Instr.Op callTargetOp = new Instr.Op(Instr.Op.Type.IMMEDIATE);
				callTargetOp.setImmediateName(callTarget);
				callInstr = new Instr("call", callTargetOp);
			} else if(addrMode != null){
				Instr.Op op = addrMode.toOp(il, ng);
				callInstr = new Instr("call", op);
			} else {
				String tempName = callTargetTile.translate(il, ng);
				callInstr = new Instr("call", new Instr.Op(Instr.Op.Type.REG, tempName));
			}
			
			il.add(callInstr);
			callInstr.setRegisterArgumentCount(regCtr);

			if (isProcedure)
				return null;

			String tempName = ng.nextTemp();			
			il.add(new Instr("mov", new Instr.Op(Instr.Op.Type.REG, tempName),
					new Instr.Op(Instr.Op.Type.REG, Configuration.RETVAL_REGISTER)));
			return tempName;
		}
	};

	public Tile match(IRNode node, Tiler tiler) {
		if (!(node instanceof IRCall))
			return null;

		IRCall call = (IRCall) node;
		List<IRExpr> params = call.params();

		List<Tile> paramTiles = new ArrayList<Tile>();
		for (int i=params.size()-1; i>=0; --i)
			paramTiles.add(tiler.tile(params.get(i)));
		
		Tile targetTile = tiler.tile(call.target());
		Tile bestTile = new CallTile(targetTile, call.isToProcedure(), paramTiles);
		int bestCost = bestTile.estimateCost();

		if (call.target() instanceof IRMem) {
			IRMem mem = (IRMem) call.target();
			List<AddrMode> modes = AddrModeHelper.findMemAddressingModes(tiler, mem);

			for (AddrMode addrMode : modes) {
				Tile tile = new CallTile(addrMode, call.isToProcedure(), paramTiles);
				int cost = tile.estimateCost();
				if (cost < bestCost) {
					bestTile = tile;
					bestCost = cost;
				}
			}
		} else if (call.target() instanceof IRName) {		
			IRName target = (IRName) call.target();
			Tile tile = new CallTile(target.name(), call.isToProcedure(), paramTiles);
			int cost = tile.estimateCost();

			if (cost < bestCost) {
				bestTile = tile;
				bestCost = cost;
			}
		}
		return bestTile;
	}
}