package xic.codegen.matchers;

import java.util.List;

import xic.codegen.AbstractTile;
import xic.codegen.Instr;
import xic.codegen.InstrList;
import xic.codegen.Matcher;
import xic.codegen.Tile;
import xic.codegen.Tiler;
import xic.codegen.matchers.AddrModeHelper.AddrMode;
import xic.ir.IRBinOp;
import xic.ir.IRConst;
import xic.ir.IRExpr;
import xic.ir.IRMem;
import xic.ir.IRMove;
import xic.ir.IRNode;
import xic.ir.IRTemp;
import xic.ir.NameGenerator;


/**
 * Handles AST fragments of the form
 * TEMP(x) <- TEMP(x) + 1
 * MEM(x)  <- MEM(x) + 1
 * TEMP(x) <- 1 + TEMP(x)
 * MEM(x)  <- 1 + MEM(x)
 * TEMP(x) <- TEMP(x) - 1
 * MEM(x)  <- MEM(x) - 1
 */
public class IncDecMatcher implements Matcher {
	protected static class IncDecTile extends AbstractTile {
		private AddrMode addrMode;
		private boolean inc;

		public IncDecTile(boolean inc, AddrMode addrMode) {
			this.inc = inc;
			this.addrMode = addrMode;
		}

		public IncDecTile(boolean inc, Tile temp) {
			this.inc = inc;
			children.add(temp);
		}

		public int localCost() {
			return 1 + (addrMode == null ? 0 : addrMode.estimateCost());
		}

		public String translate(InstrList il, NameGenerator ng) {
			String mnemonic = inc ? "inc" : "dec";
			if (addrMode != null)
				il.add(new Instr(mnemonic, addrMode.toOp(il, ng)));
			else
				il.add(new Instr(mnemonic, 
				new Instr.Op(Instr.Op.Type.REG, children.get(0).translate(il, ng))));
			return null;
		}
	};

	public Tile match(IRNode node, Tiler tiler) {
		if (!(node instanceof IRMove))
			return null;

		IRMove move = (IRMove) node;
		if (!(move.expr() instanceof IRBinOp))
			return null;

		IRBinOp binOp = (IRBinOp) move.expr();
		IRBinOp.OpType opType = binOp.opType();
		boolean inc;
		IRExpr expr;

		if (binOp.left() instanceof IRConst 
				&& ((IRConst) binOp.left()).value() == 1
				&& binOp.right().equalsTree(move.target())
				&& opType == IRBinOp.OpType.ADD) {
			inc = true;
			expr = binOp.right();
		} else if (binOp.right() instanceof IRConst 
				&& ((IRConst) binOp.right()).value() == 1
				&& binOp.left().equalsTree(move.target())
				&& (opType == IRBinOp.OpType.ADD
					|| opType == IRBinOp.OpType.SUB)) {
			inc = (opType == IRBinOp.OpType.ADD);
			expr = binOp.left();
		} else {
			return null;
		}

		if (expr instanceof IRTemp) {
			return new IncDecTile(inc, tiler.tile(expr));
		} else if (expr instanceof IRMem) {
			Tile bestTile = null;
			int bestCost = Integer.MAX_VALUE;
			List<AddrMode> modes = AddrModeHelper.findMemAddressingModes(tiler, (IRMem) expr);

			for (AddrMode addrMode : modes) {
				Tile tile = new IncDecTile(inc, addrMode);
				int cost = tile.estimateCost();
				if (cost < bestCost) {
					bestTile = tile;
					bestCost = cost;
				}
			}
			return bestTile;
		} else {
			return null;
		}
	}
}
