package xic.ir;

import java.util.Arrays;
import java.util.Set;

import xic.InternalCompilerException;
import xic.ast.ASTNode;
import xic.ast.Copyable;
import xic.visitors.NodeVisitor;

import edu.cornell.cs.cs4120.util.VisualizableTreeNode;

public class IRBinOp extends IRExpr {
	public enum OpType {
		ADD, SUB, MUL, DIV, MOD, AND, OR, LOGAND, 
		LOGOR, XOR, LSHIFT, RSHIFT, AR_LSHIFT, 
		AR_RSHIFT, EQ, NEQ, LT, GT, LE, GE, 
		ULT, UGT, ULE, UGE
	};

	private OpType type;
	private IRExpr left, right;

	public IRBinOp(OpType type, IRExpr left, IRExpr right) {
		this.type = type;
		this.left = left;
		this.right = right;
	}

	public OpType opType() {
		return type;
	}

	public IRExpr left() {
		return left;
	}

	public IRExpr right() {
		return right;
	}

	public Iterable<VisualizableTreeNode> children() {
		return Arrays.asList(new VisualizableTreeNode[] { left, right });
	}

	public ASTNode visitChildren(NodeVisitor v) {
		IRExpr left = (IRExpr) v.visit(this, this.left);
		IRExpr right = (IRExpr) v.visit(this, this.right);

		if (left != this.left || right != this.right)
			return new IRBinOp(type, left, right);

		return this;
	}

	public IRExpr negate() {
		IRBinOp expr = (IRBinOp) copy();
		switch (type) {
		case LT:
			expr.type = OpType.GE;
			break;
		case GT:
			expr.type = OpType.LE;
			break;
		case LE:
			expr.type = OpType.GT;
			break;
		case GE:
			expr.type = OpType.LT;
			break;
		case ULT:
			expr.type = OpType.UGE;
			break;
		case UGT:
			expr.type = OpType.ULE;
			break;
		case ULE:
			expr.type = OpType.UGT;
			break;
		case UGE:
			expr.type = OpType.ULT;
			break;
		case EQ:
			expr.type = OpType.NEQ;
			break;
		case NEQ:
			expr.type = OpType.EQ;
			break;
		default:
			return super.negate();
		}
		return expr;
	}

	public IRBinOp swapArgs() {
		IRBinOp expr = (IRBinOp) copy();
		expr.left = right;
		expr.right = left;
		switch (type) {
		case LT:
			expr.type = OpType.GT;
			break;
		case GT:
			expr.type = OpType.LT;
			break;
		case LE:
			expr.type = OpType.GE;
			break;
		case GE:
			expr.type = OpType.LE;
			break;
		case ULT:
			expr.type = OpType.UGT;
			break;
		case UGT:
			expr.type = OpType.ULT;
			break;
		case ULE:
			expr.type = OpType.UGE;
			break;
		case UGE:
			expr.type = OpType.ULE;
			break;
		case EQ:
			expr.type = OpType.EQ;
			break;
		case NEQ:
			expr.type = OpType.NEQ;
			break;
		case ADD:
			expr.type = OpType.ADD;
			break;
		case MUL:
			expr.type = OpType.MUL;
			break;
		case AND:
			expr.type = OpType.AND;
			break;
		case OR:
			expr.type = OpType.OR;
			break;
		case LOGAND:
			expr.type = OpType.LOGAND;
			break;
		case LOGOR:
			expr.type = OpType.LOGOR;
			break;
		default:
			return null;
		}
		return expr;
	}

	public boolean isComparison() {
		switch (type) {
			case EQ:
			case NEQ:
			case LT:
			case GT:
			case LE:
			case GE:
			case ULT:
			case UGT:
			case ULE:
			case UGE:
				return true;
			default:
				return false;
		}
	}

	public String label() {
		switch (type) {
		case ADD:
			return "ADD";
		case SUB:
			return "SUB";
		case MUL:
			return "MUL";
		case DIV:
			return "DIV";
		case MOD:
			return "MOD";
		case AND:
			return "AND";
		case OR:
			return "OR";
		case LOGAND:
			return "LOGAND";
		case LOGOR:
			return "LOGOR";
		case XOR:
			return "XOR";
		case LSHIFT:
			return "LSHIFT";
		case RSHIFT:
			return "RSHIFT";
		case AR_LSHIFT:
			return "AR_LSHIFT";
		case AR_RSHIFT:
			return "AR_RSHIFT";
		case EQ:
			return "EQ";
		case NEQ:
			return "NEQ";
		case LT:
			return "LT";
		case GT:
			return "GT";
		case LE:
			return "LE";
		case GE:
			return "GE";
		case ULT:
			return "ULT";
		case UGT:
			return "UGT";
		case ULE:
			return "ULE";
		case UGE:
			return "UGE";
		default:
			return "invalid";
		}
	}

	public boolean containsCalls() {
		return left.containsCalls() || right.containsCalls();
	}

	public boolean containsMemReads() {
		return left.containsMemReads() || right.containsMemReads();
	}

	public boolean containsMemWrites() {
		return left.containsMemWrites() || right.containsMemWrites();
	}

	public void findReadTemps(Set<String> tempNames) {
		left.findReadTemps(tempNames);
		right.findReadTemps(tempNames);
	}

	public void findWrittenTemps(Set<String> tempNames) {
		left.findWrittenTemps(tempNames);
		right.findWrittenTemps(tempNames);
	}

	public int computeMaximumCallArguments() {
		int l = left.computeMaximumCallArguments();
		int r = right.computeMaximumCallArguments();
		return Math.max(l, r);
	}

	public int computeMaximumCallResults() {
		int l = left.computeMaximumCallResults();
		int r = right.computeMaximumCallResults();
		return Math.max(l, r);
	}

	public int nodeCount() {
		return 1 + left.nodeCount() + right.nodeCount();
	}

	public Copyable copy() {
		return new IRBinOp(type, left, right);
	}

	public Copyable deepCopy() {
		return new IRBinOp(type, (IRExpr) left.deepCopy(), (IRExpr) right
				.deepCopy());
	}

	public boolean equalsTree(Object object) {
		if (!(object instanceof IRBinOp))
			return false;
		IRBinOp other = (IRBinOp) object;
		return other.left.equalsTree(left) && other.right.equalsTree(right);
	}

	public int treeHashCode() {
		int code = 17;
		code = 37 * code + type.hashCode();
		code = 37 * code + left.treeHashCode();
		code = 37 * code + right.treeHashCode();
		return code;
	}

	// eval the value based on op type
	public static long eval(OpType opType, long leftVal, long rightVal) {
		switch (opType) {
		case ADD:
			return leftVal + rightVal;
		case AND:
		case LOGAND:
			return leftVal & rightVal;
		case DIV:
			if (rightVal == 0)
				throw new InternalCompilerException("Division by zero " +
					"during constant folding -- please fix your program!");
			return leftVal / rightVal;
		case EQ:
			return leftVal == rightVal ? 1 : 0;
		case GE:
			return leftVal >= rightVal ? 1 : 0;
		case GT:
			return leftVal > rightVal ? 1 : 0;
		case LE:
			return leftVal <= rightVal ? 1 : 0;
		case AR_LSHIFT:
			return leftVal << rightVal;
		case LSHIFT:
			return leftVal << rightVal;
		case LT:
			return leftVal < rightVal ? 1 : 0;
		case MOD:
			if (rightVal == 0)
				throw new InternalCompilerException("Division by zero " +
					"during constant folding -- please fix your program!");
			return leftVal % rightVal;
		case MUL:
			return leftVal * rightVal;
		case NEQ:
			return leftVal != rightVal ? 1 : 0;
		case OR:
		case LOGOR:
			return leftVal | rightVal;
		case AR_RSHIFT:
			return leftVal >> rightVal;
		case RSHIFT:
			return leftVal >>> rightVal;
		case SUB:
			return leftVal - rightVal;
		case XOR:
			return leftVal ^ rightVal;
		case UGT:
			return ((leftVal > rightVal) 
				^ ((leftVal < 0) != (rightVal < 0))) ? 1 : 0;
		case ULT:
			return ((leftVal < rightVal) 
				^ ((leftVal < 0) != (rightVal < 0))) ? 1 : 0;
		case ULE:
			return ((leftVal == rightVal) || ((leftVal < rightVal) 
					^ ((leftVal < 0) != (rightVal < 0)))) ? 1 : 0;
		case UGE:
			return ((leftVal == rightVal) || ((leftVal > rightVal) 
					^ ((leftVal < 0) != (rightVal < 0)))) ? 1 : 0;
		default:
			throw new InternalCompilerException("Unknow op type "
					+ opType.toString());
		}
	}
}
