package xic.ir;

import java.util.Arrays;
import java.util.Set;

import xic.InternalCompilerException;
import xic.ast.ASTNode;
import xic.ast.Copyable;
import xic.codegen.RegAlloc;
import xic.visitors.NodeVisitor;
import edu.cornell.cs.cs4120.util.VisualizableTreeNode;

/**
 * Denotes either a fetch or a store, depending on usage. 
 */
public class IRMem extends IRExpr {
	public enum MemType {
		NORMAL,
		IMMUTABLE
	};

	private IRExpr expr;
	private MemType memType;

	public IRMem(IRExpr expr) {
		this.expr = expr;
		this.memType = MemType.NORMAL;
	}

	public IRMem(IRExpr expr, MemType memType) {
		this.expr = expr;
		this.memType = memType;
	}

	public IRExpr expr() {
		return expr;
	}
	
	public MemType memType() {
		return memType;
	}
	
	public Iterable<VisualizableTreeNode> children() {
		return Arrays.asList(new VisualizableTreeNode[] {
			expr
		});
	}

    public ASTNode visitChildren(NodeVisitor v) {
    	IRExpr expr = (IRExpr) v.visit(this, this.expr);

		if (expr != this.expr)
			return new IRMem(expr, memType);

		return this;
	}

    public boolean containsCalls() {
    	return expr.containsCalls();
    }
    
    public boolean containsMemReads() {
    	if (memType == MemType.IMMUTABLE) // treat as constant data
    		return false;
    	return true;
    }

    public boolean containsMemWrites() {
    	return false;
    }

    public void findReadTemps(Set<String> tempNames) {
    	expr.findReadTemps(tempNames);
    	
    	if (memType != MemType.IMMUTABLE)
    		tempNames.add(RegAlloc.alias(this));
    }

    public void findWrittenTemps(Set<String> tempNames) {
    	expr.findWrittenTemps(tempNames);
    }

    public int nodeCount() {
    	return 1 + expr.nodeCount();
    }

	public String label() {
		switch (memType) {
			case IMMUTABLE:
				return "MEM_I";
			case NORMAL:
				return "MEM";
			default:
				throw new InternalCompilerException("Unknown array type!");
		}
	}
	
	public int computeMaximumCallArguments() {
		return expr.computeMaximumCallArguments();
	}

	public int computeMaximumCallResults() {
		return expr.computeMaximumCallResults();
	}

	public Copyable copy() {
		return new IRMem(expr, memType);
	}

	public Copyable deepCopy() {
		return new IRMem((IRExpr) expr.deepCopy(), memType);
	}	

	public boolean equalsTree(Object object) {
		if (!(object instanceof IRMem))
			return false;
		return ((IRMem) object).expr.equalsTree(expr);
	}
	
	public int treeHashCode() {
		return expr.treeHashCode();
	}
}
