package xic.ir;

import java.util.Arrays;
import java.util.Set;

import xic.InternalCompilerException;
import xic.ast.ASTNode;
import xic.ast.Copyable;
import xic.codegen.RegAlloc;
import xic.ir.IRMem.MemType;
import xic.visitors.NodeVisitor;
import edu.cornell.cs.cs4120.util.VisualizableTreeNode;

public class IRMove extends IRStmt {
	private IRExpr target;
	private IRExpr expr;
	
	public IRMove(IRExpr target, IRExpr expr) {
		this.target = target;
		this.expr = expr;
	}

	public IRExpr target() {
		return target;
	}

	public IRExpr expr() {
		return expr;
	}

	public Iterable<VisualizableTreeNode> children() {
		return Arrays.asList(new VisualizableTreeNode[] {
			target, expr
		});
	}

    public ASTNode visitChildren(NodeVisitor v) {
    	IRExpr target = (IRExpr) v.visit(this, this.target);
    	IRExpr expr = (IRExpr) v.visit(this, this.expr);

		if (target != this.target || expr != this.expr)
			return new IRMove(target, expr);

		return this;
	}
    
    public boolean containsCalls() {
    	return target.containsCalls() || expr.containsCalls();
    }

    public boolean containsMemReads() {
    	if (target instanceof IRMem)
    		return expr.containsMemReads() || ((IRMem) target).expr().containsMemReads();
    	else
    		return target.containsMemReads() || expr.containsMemReads();
    }

    public boolean containsMemWrites() {
    	if (target instanceof IRMem)
    		return true;
    	else
    		return target.containsMemWrites() || expr.containsMemWrites();
    }

    public void findReadTemps(Set<String> tempNames) {
    	if (target instanceof IRTemp) {
    		// nothing
    	} else {
    		target.findReadTemps(tempNames);
    	}
    	expr.findReadTemps(tempNames);
    }

    public void findWrittenTemps(Set<String> tempNames) {
    	if (target instanceof IRTemp) {
    		tempNames.add(((IRTemp) target).name());
    	} else if (target instanceof IRMem) {
    		IRMem mem = (IRMem) target;
    		if (mem.memType() == MemType.IMMUTABLE)
    			throw new InternalCompilerException("Encountered an " +
    				"operation that writes to memory marked as immutable!");
        	tempNames.add(RegAlloc.alias(mem));
    		target.findWrittenTemps(tempNames);
    	} else {
    		target.findWrittenTemps(tempNames);
    	}
    	expr.findWrittenTemps(tempNames);
    }
    
    public int nodeCount() {
    	return 1 + target.nodeCount() + expr.nodeCount();
    }

	public int computeMaximumCallArguments() {
		int l = target.computeMaximumCallArguments();
		int r = expr.computeMaximumCallArguments();
		return Math.max(l, r);
	}

	public int computeMaximumCallResults() {
		int l = target.computeMaximumCallResults();
		int r = expr.computeMaximumCallResults();
		return Math.max(l, r);
	}

	public String label() {
		return "MOVE";
	}

	public Copyable copy() {
		return new IRMove(target, expr);
	}
	
	public Copyable deepCopy() {
		return new IRMove((IRExpr) target.deepCopy(),
					(IRExpr) expr.deepCopy());
	}
	
	public boolean equalsTree(Object object) {
		if (!(object instanceof IRMove))
			return false;
		IRMove other = (IRMove) object;
		return other.expr.equalsTree(expr) 
			&& other.target.equalsTree(target);
	}
	
	public int treeHashCode() {
		return 17 + target.treeHashCode() + expr.treeHashCode() * 37;
	}
}
