package xic;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.junit.Test;

import xic.Configuration;
import xic.Statistics;
import xic.analysis.AnticipatedExpressions;
import xic.analysis.AvailableExpressions;
import xic.analysis.CFGNode;
import xic.analysis.Delayedness;
import xic.analysis.DotGenerator;
import xic.analysis.EarliestExpressions;
import xic.analysis.Isolated;
import xic.analysis.Latestness;
import xic.analysis.PartialRedundancy;
import xic.analysis.WebGenerator;
import xic.analysis.WebGenerator.CoalesceCriteria;
import xic.analysis.WebGenerator.Web;
import xic.codegen.AsmDirective;
import xic.codegen.Instr;
import xic.ir.IRBinOp;
import xic.ir.IRCJump;
import xic.ir.IRConst;
import xic.ir.IRJump;
import xic.ir.IRLabel;
import xic.ir.IRMove;
import xic.ir.IRName;
import xic.ir.IRNode;
import xic.ir.IRStmt;
import xic.ir.IRTemp;
import xic.ir.NameGenerator;
import xic.ir.IRBinOp.OpType;


public class AnalysisTest {
	@Test
	public void testWebGenerator() {
		/* Verifies the web generation example in Figure 16.5, Muchnick. */
		CFGNode root = new CFGNode(new AsmDirective("entry"));

		CFGNode defX1 = new CFGNode(new Instr("mov",
				new Instr.Op(Instr.Op.Type.REG, "x"),
				new Instr.Op("B2_defX")
			));
		CFGNode defY1 = new CFGNode(new Instr("mov",
				new Instr.Op(Instr.Op.Type.REG, "y"),
				new Instr.Op("B2_defY")
			));
		root.addEdgeTo(defX1);
		defX1.addEdgeTo(defY1);
		
		CFGNode useX1 = new CFGNode(new Instr("cmp", 
				new Instr.Op(Instr.Op.Type.REG, "x"),
				new Instr.Op("B4_useX")
			));
		CFGNode useY1 = new CFGNode(new Instr("cmp",
				new Instr.Op(Instr.Op.Type.REG, "y"),
				new Instr.Op("B4_useY")
			));
		defY1.addEdgeTo(useX1);
		useX1.addEdgeTo(useY1);

		CFGNode useX2 = new CFGNode(new Instr("cmp",
				new Instr.Op(Instr.Op.Type.REG, "x"),
				new Instr.Op("B5_useX")
			));
		CFGNode defX2 = new CFGNode(new Instr("mov",
				new Instr.Op(Instr.Op.Type.REG, "x"),
				new Instr.Op("B5_defX")
			));
		CFGNode useX3 = new CFGNode(new Instr("cmp",
				new Instr.Op(Instr.Op.Type.REG, "x"),
				new Instr.Op("B6_useX")
			));

		defY1.addEdgeTo(useX2);
		useX2.addEdgeTo(defX2);
		defX2.addEdgeTo(useX3);
		
		CFGNode defY2 = new CFGNode(new Instr("mov",
				new Instr.Op(Instr.Op.Type.REG, "y"),
				new Instr.Op("B1_defY")
			));
		CFGNode defX3 = new CFGNode(new Instr("mov",
				new Instr.Op(Instr.Op.Type.REG, "x"),
				new Instr.Op("B3_defX")
			));
		CFGNode useY3 = new CFGNode(new Instr("cmp",
				new Instr.Op(Instr.Op.Type.REG, "y"),
				new Instr.Op("B3_useY")
			));
		root.addEdgeTo(defY2);
		defY2.addEdgeTo(defX3);
		defX3.addEdgeTo(useY3);
		useY3.addEdgeTo(useX2);
		
		
		CFGNode exit = new CFGNode(new AsmDirective("exit"));
		useY1.addEdgeTo(exit);
		useX3.addEdgeTo(exit);

		Set<Web> webs = WebGenerator.run("", new Statistics(), 0, root.flatten(),
				CoalesceCriteria.COALESCE_BRIGGS);

		for (Iterator<Web> it = webs.iterator(); it.hasNext(); ) {
			if (Configuration.REG_NAMES.contains(it.next().toString()))
				it.remove();
		}

		assertEquals(4, webs.size());
		int seen = 0;
		Set<String> defNames = new HashSet<String>();
		Set<String> useNames = new HashSet<String>();
		
		for (Web w : webs) {
			defNames.clear(); useNames.clear();

			for (CFGNode n : w.defs())
				defNames.add(((Instr) n.value()).getOp2().getImmediateName());
			for (CFGNode n : w.uses())
				useNames.add(((Instr) n.value()).getOp2().getImmediateName());
			if (w.varNames().contains("x") && w.defs().size() == 2 && w.uses().size() == 2
					&& defNames.contains("B2_defX") && defNames.contains("B3_defX")
					&& useNames.contains("B5_useX") && useNames.contains("B5_useX"))
				seen |= 0x1;
			if (w.varNames().contains("x") && w.defs().size() == 1 && w.uses().size() == 1
					&& defNames.contains("B5_defX") && useNames.contains("B6_useX"))
				seen |= 0x2;
			if (w.varNames().contains("y") && w.defs().size() == 1 && w.uses().size() == 1
					&& defNames.contains("B2_defY") && useNames.contains("B4_useY"))
				seen |= 0x4;
			if (w.varNames().contains("y") && w.defs().size() == 1 && w.uses().size() == 1
					&& defNames.contains("B1_defY") && useNames.contains("B3_useY"))
				seen |= 0x8;
		}
		if (seen != 0xf)
			fail();
	}

	@Test
	public void testPartialRedundancy() {
		CFGNode head = new CFGNode("entry");
		
		IRStmt ir1 = new IRMove(new IRTemp("z"), 
				new IRBinOp(OpType.MUL, new IRTemp("a"), new IRTemp("xx")));
		CFGNode n1 = new CFGNode(ir1);
		head.addEdgeTo(n1);
		
		IRStmt ir2 = new IRCJump(new IRBinOp(OpType.GT, 
								new IRTemp("x"), new IRConst(3)), 
								"B2", "B3");
		CFGNode n2 = new CFGNode(ir2);
		n1.addEdgeTo(n2);
		
		IRStmt ir3 = new IRLabel("B2");
		CFGNode n3 = new CFGNode(ir3);
		n2.addEdgeTo(n3);
		
		IRStmt ir4 = new IRMove(new IRTemp("a"), 
				new IRBinOp(OpType.MUL, new IRTemp("x"),new IRTemp("y")));
		CFGNode n4 = new CFGNode(ir4);
		n3.addEdgeTo(n4);

		
		IRStmt ir5 = new IRCJump(new IRBinOp(OpType.LT, 
				new IRTemp("y"), new IRConst(5)), 
				"B4", "B2a");
		CFGNode n5 = new CFGNode(ir5);
		n4.addEdgeTo(n5);
		
		IRStmt ir6 = new IRLabel("B3");
		CFGNode n6 = new CFGNode(ir6);
		n2.addEdgeTo(n6);
		
		IRStmt ir7 = new IRCJump(new IRBinOp(OpType.LT, 
				new IRTemp("z"), new IRConst(7)), 
				"B3a", "B6");
		CFGNode n7 = new CFGNode(ir7);
		n6.addEdgeTo(n7);
		
		IRStmt ir8 = new IRLabel("B4");
		CFGNode n8 = new CFGNode(ir8);
		n5.addEdgeTo(n8);
		
		IRStmt ir9 = new IRMove(new IRTemp("b"), 
				new IRBinOp(OpType.MUL, new IRTemp("x"),new IRTemp("y")));
		CFGNode n9 = new CFGNode(ir9);
		n8.addEdgeTo(n9);

		IRStmt ir10 = new IRJump(new IRName("exit"));
		CFGNode n10 = new CFGNode(ir10);
		n9.addEdgeTo(n10);
		
		CFGNode n11 = new CFGNode(new IRLabel("exit"));
		n10.addEdgeTo(n11);
		CFGNode exit = new CFGNode("exit");
		n11.addEdgeTo(exit);
		
		IRStmt ir12 = new IRLabel("B5");
		CFGNode n12 = new CFGNode(ir12);
		
		CFGNode B3a = new CFGNode(new IRLabel("B3a"));
		n7.addEdgeTo(B3a);
		CFGNode B2a = new CFGNode(new IRLabel("B2a"));
		n5.addEdgeTo(B2a);
		
		B3a.addEdgeTo(n12);
		B2a.addEdgeTo(n12);

		IRStmt ir13 = new IRMove(new IRTemp("c"), 
				new IRBinOp(OpType.MUL, new IRTemp("x"),new IRTemp("y")));
		CFGNode n13 = new CFGNode(ir13);
		n12.addEdgeTo(n13);
		
		IRStmt ir14 = new IRJump(new IRName("exit"));
		CFGNode n14 = new CFGNode(ir14);
		n13.addEdgeTo(n14);
		n14.addEdgeTo(n11);
		
		IRStmt ir15 = new IRLabel("B6");
		CFGNode n15 = new CFGNode(ir15);
		n7.addEdgeTo(n15);

		IRStmt ir16 = new IRJump(new IRName("exit"));
		CFGNode n16 = new CFGNode(ir16);
		n15.addEdgeTo(n16);
		n16.addEdgeTo(n11);

	    Set<CFGNode> cfgNodes = head.flatten();

	    AvailableExpressions ae = new AvailableExpressions(cfgNodes);
		ae.solve();

		DotGenerator dg = new DotGenerator();
	    dg.setDataFlowAnalysis(ae, cfgNodes);

		IRNode aplusone = ae.eval(n1).iterator().next();
		IRNode xgt3 = ae.eval(n2).iterator().next();
		IRNode xtimesy = ae.eval(n4).iterator().next();
		IRNode ylt5 = ae.eval(n5).iterator().next();
		IRNode zlt7 = ae.eval(n7).iterator().next();

		assertTrue(ae.in(head).size()==0);
		assertTrue(ae.out(head).size()==0);
		
		assertTrue(ae.in(n1).size()==0);
		assertTrue(ae.out(n1).contains(aplusone));
		
		assertTrue(ae.in(n2).contains(aplusone));
		assertTrue(ae.out(n2).contains(aplusone));
		assertTrue(ae.out(n2).contains(xgt3));

		assertTrue(ae.in(n4).contains(aplusone));
		assertTrue(ae.in(n4).contains(xgt3));
		assertFalse(ae.out(n4).contains(aplusone));
		assertTrue(ae.out(n4).contains(xtimesy));
		assertTrue(ae.out(n4).contains(xgt3));
		
		assertFalse(ae.in(n12).contains(xtimesy));
		assertFalse(ae.in(n12).contains(aplusone));
		assertTrue(ae.in(n12).contains(xgt3));
		assertTrue(ae.out(n12).contains(xgt3));		

		assertFalse(ae.in(n11).contains(xtimesy));
		assertFalse(ae.in(n11).contains(aplusone));
		assertTrue(ae.in(n11).contains(xgt3));
		assertFalse(ae.in(n11).contains(zlt7));
		assertFalse(ae.in(n11).contains(ylt5));	

	    AnticipatedExpressions ante = new AnticipatedExpressions(cfgNodes);
		ante.solve();

		assertTrue(ante.in(head).contains(aplusone));
		assertTrue(ante.out(head).contains(aplusone));
		
		assertTrue(ante.in(n1).contains(aplusone));
		assertTrue(ante.in(n1).contains(xgt3));
		assertFalse(ante.out(n1).contains(aplusone));
		assertTrue(ante.out(n1).contains(xgt3));
		
		assertTrue(ante.in(n2).contains(xgt3));
		assertFalse(ante.out(n2).contains(xgt3));

		assertTrue(ante.in(n4).contains(xtimesy));
	    EarliestExpressions early = new EarliestExpressions(ante, cfgNodes);
		early.solve();
		assertTrue(early.in(head).contains(aplusone));
		assertTrue(early.in(head).contains(xtimesy));
		assertTrue(early.in(head).contains(xgt3));
		assertTrue(early.in(head).contains(ylt5));
		assertTrue(early.in(head).contains(zlt7));
		
		assertTrue(early.out(head).contains(xtimesy));
		assertFalse(early.out(head).contains(aplusone));
		assertFalse(early.out(head).contains(xgt3));
		assertTrue(early.out(head).contains(ylt5));
		assertTrue(early.out(head).contains(zlt7));

		assertTrue(early.out(n1).contains(xtimesy));
		assertFalse(early.out(n1).contains(aplusone));
		assertFalse(early.out(n1).contains(xgt3));
		assertTrue(early.out(n1).contains(ylt5));
		assertTrue(early.out(n1).contains(zlt7));
		
		assertTrue(early.in(n8).contains(aplusone));
		assertTrue(early.out(n8).contains(aplusone));

		Delayedness del = new Delayedness(cfgNodes, head, ante, early);
		del.solve();
		
		assertTrue(del.in(head).size() == 2);
		assertTrue(del.in(head).contains(aplusone));
		assertTrue(del.in(head).contains(xgt3));
		assertTrue(del.out(head).size() == 2);
		assertTrue(del.out(head).contains(aplusone));
		assertTrue(del.out(head).contains(xgt3));

		assertTrue(del.in(n1).size() == 2);
		assertTrue(del.in(n1).contains(aplusone));
		assertTrue(del.in(n1).contains(xgt3));
		assertTrue(del.out(n1).size() == 1);
		assertTrue(del.out(n1).contains(xgt3));
		
		assertTrue(del.in(n2).size() == 1);
		assertTrue(del.in(n2).contains(xgt3));
		assertTrue(del.out(n2).size() == 0);
		
		assertTrue(del.in(n3).size() == 2);
		assertTrue(del.in(n3).contains(xtimesy));
		assertTrue(del.in(n3).contains(ylt5));
		assertTrue(del.out(n3).size() == 2);
		assertTrue(del.out(n3).contains(xtimesy));
		assertTrue(del.out(n3).contains(ylt5));
		
		assertTrue(del.in(n4).size() == 2);
		assertTrue(del.in(n4).contains(xtimesy));
		assertTrue(del.in(n4).contains(ylt5));
		assertTrue(del.out(n4).size() == 1);
		assertTrue(del.out(n4).contains(ylt5));
		
		assertTrue(del.in(n5).size() == 1);
		assertTrue(del.in(n5).contains(ylt5));
		assertTrue(del.out(n5).size() == 0);

		assertTrue(del.in(n8).size() == 0);
		assertTrue(del.out(n8).size() == 0);
		assertTrue(del.in(n9).size() == 0);
		assertTrue(del.out(n9).size() == 0);
		assertTrue(del.in(n10).size() == 0);
		assertTrue(del.out(n10).size() == 0);
		assertTrue(del.in(n11).size() == 0);
		assertTrue(del.out(n11).size() == 0);
		
		assertTrue(del.in(n12).size() == 0);
		assertTrue(del.out(n12).size() == 0);

		assertTrue(del.in(n6).size() == 1);
		assertTrue(del.in(n6).contains(zlt7));
		assertTrue(del.out(n6).size() == 1);
		assertTrue(del.out(n6).contains(zlt7));

		assertTrue(del.in(n7).size() == 1);
		assertTrue(del.in(n7).contains(zlt7));
		assertTrue(del.out(n7).size() == 0);
		assertTrue(del.in(n15).size() == 0);
		assertTrue(del.out(n15).size() == 0);
		assertTrue(del.in(n16).size() == 0);
		assertTrue(del.out(n16).size() == 0);

		assertTrue(del.in(B2a).size() == 0);
		assertTrue(del.out(B2a).size() == 0);
		assertTrue(del.in(B3a).size() == 1);
		assertTrue(del.in(B3a).contains(xtimesy));
		assertTrue(del.out(B3a).size() == 1);
		assertTrue(del.out(B3a).contains(xtimesy));
		
		
		Latestness lt = new Latestness(cfgNodes, ante, del);
		assertTrue(lt.in(head).size() == 0);
		assertTrue(lt.in(n1).size() == 1);
		assertTrue(lt.in(n1).contains(aplusone));
		assertTrue(lt.in(n2).size() == 1);
		assertTrue(lt.in(n2).contains(xgt3));

		assertTrue(lt.in(head).size() == 0);
		assertTrue(lt.in(n1).size() == 1);
		assertTrue(lt.in(n1).contains(aplusone));
		assertTrue(lt.in(n2).size() == 1);
		assertTrue(lt.in(n2).contains(xgt3));
		assertTrue(lt.in(n3).size() == 0);
		assertTrue(lt.in(n4).size() == 1);
		assertTrue(lt.in(n4).contains(xtimesy));
		assertTrue(lt.in(n8).size() == 0);
		assertTrue(lt.in(n9).size() == 0);
		assertTrue(lt.in(n10).size() == 0);
		assertTrue(lt.in(n14).size() == 0);
		assertTrue(lt.in(n13).size() == 0);
		assertTrue(lt.in(n12).size() == 0);
		assertTrue(lt.in(B2a).size() == 0);
		assertTrue(lt.in(B3a).size() == 1);
		assertTrue(lt.in(B3a).contains(xtimesy));

		assertTrue(lt.in(n6).size() == 0);
		assertTrue(lt.in(n7).size() == 1);
		assertTrue(lt.in(n7).contains(zlt7));
		assertTrue(lt.in(n15).size() == 0);
		assertTrue(lt.in(n16).size() == 0);
		assertTrue(lt.in(n11).size() == 0);
		
		Isolated iso = new Isolated(cfgNodes, head, ante, lt);
		iso.solve();
		assertTrue(iso.in(head).size() == 2);
		assertTrue(iso.in(head).contains(aplusone));
		assertTrue(iso.in(head).contains(xgt3));

		assertTrue(iso.in(n1).size() == 2);
		assertTrue(iso.in(n1).contains(aplusone));
		assertTrue(iso.in(n1).contains(xgt3));
		
		assertTrue(iso.out(n1).size() == 1);
		assertFalse(iso.out(n1).contains(aplusone));
		assertTrue(iso.out(n1).contains(xgt3));

		assertTrue(iso.in(n2).size() == 1);
		assertTrue(iso.in(n2).contains(xgt3));
		assertTrue(iso.out(n2).size() == 0);

		assertTrue(iso.in(n3).size() == 2);
		assertTrue(iso.in(n3).contains(xtimesy));
		assertTrue(iso.in(n3).contains(ylt5));

		assertTrue(iso.in(B3a).size() == 1);
		assertTrue(iso.in(B3a).contains(xtimesy));

		assertTrue(iso.in(n6).size() == 1);
		assertTrue(iso.in(n6).contains(zlt7));

		assertTrue(iso.in(n7).size() == 1);
		assertTrue(iso.in(n7).contains(zlt7));
		assertTrue(iso.out(n7).size() == 0);

		assertTrue(iso.in(n15).size() == 0);
		assertTrue(iso.in(n16).size() == 0);
		assertTrue(iso.in(n11).size() == 0);

		PartialRedundancy pre = new PartialRedundancy(new Statistics(), new NameGenerator());
		pre.analyze("foo", head);

		Map<String, CFGNode> map = new HashMap<String,CFGNode>();
		map.put("foo",head);
		/*
			DotGenerator dot = new DotGenerator();
			for (CFGNode node : cfgNodes) {
				dot.getIn().put(node, String.valueOf(pre.opt(node)));
				dot.getOut().put(node, String.valueOf(pre.redn(node)));
			}
			System.out.println(dot.generate(head.flatten()));
		*/	
		assertTrue(pre.opt(head).size() == 0);
		assertTrue(pre.redn(head).size() == 0);

	
		assertTrue(pre.opt(n1).size()==1);
		assertTrue(pre.opt(n1).contains(aplusone));
		assertTrue(pre.opt(n4).size()==1);
		assertTrue(pre.opt(n4).contains(xtimesy));
		assertTrue(pre.opt(B3a).size()==1);
		assertTrue(pre.opt(B3a).contains(xtimesy));
	
		assertTrue(pre.opt(B2a).size()==0);
		assertTrue(pre.opt(n9).size()==0);
		assertTrue(pre.opt(n8).size()==0);
		assertTrue(pre.opt(n10).size()==0);

		assertTrue(pre.redn(n9).size()==1);
		assertTrue(pre.redn(n9).contains(xtimesy));
		assertTrue(pre.redn(n13).size()==1);
		assertTrue(pre.redn(n13).contains(xtimesy));
	}
}
