package xic;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.io.BufferedReader;
import java.io.CharArrayReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Arrays;
import java.util.List;

import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameters;

import xic.Configuration.Architecture;
import xic.Configuration.AssemblySyntax;
import xic.Configuration.Target;


@RunWith(Parameterized.class)
public class ExecutableTest {
	@Parameters
	public static List<Boolean[]> parameters() {
		return Arrays.asList(new Boolean[][] {
				{ false, false, false }, /* 32 bit, without optimizations */
				{ false, true, false },  /* 32 bit, with optimizations */
				{ true, false, true },   /* 64 bit, without optimizations */
				{ true, true, true }     /* 64 bit, with optimizations, use AT&T syntax :) */
		});
	};
	private boolean optimize;

	public ExecutableTest(boolean x86_64, boolean optimize, boolean att) {
		Configuration.setArchitecture(x86_64 ? Architecture.X86_64 : Architecture.I386);
		Configuration.setSyntax(att ? AssemblySyntax.ATT_SYNTAX : AssemblySyntax.INTEL_SYNTAX); 
		Configuration.setTarget(Configuration.detectTarget());
		this.optimize = optimize;
	}

	public static String getIotaRuntimeDirectory() {
		try {
			/* Whose computer is this? */
			String hostName = InetAddress.getLocalHost().getHostName()
					.toLowerCase();

			if (hostName.equals("maxwell") 
					|| hostName.equals("wenzel5132up")) {
					if (Configuration.ARCH == Architecture.I386)
						return "/home/wenzel/xic/runtime/x86";
					else
						return "/home/wenzel/xic/runtime/x86_64";
			} else if (hostName.equals("feynman")) {
					if (Configuration.ARCH == Architecture.I386)
						return "/Users/wenzel/xic/runtime/x86";
					else
						return "/Users/wenzel/xic/runtime/x86_64";

			} else if (hostName.equals("wenzel-pc")) {
				return "C:/Users/Wenzel/Desktop/xic/runtime/x86_64";
			} else {
				fail("Don't know where to find the runtime on your machine - "
						+ "add a line to ExecutableTest.java matching the host"
						+ "name '" + hostName + "'");
			}
		} catch (UnknownHostException e) {
			fail(e.getMessage());
		}
		return null;
	}
	
	public static String getIotaIncludeDirectory() {
		return getIotaRuntimeDirectory() + "/include";
	}

	private String getCygwinPath() {
		try {
			String hostName = InetAddress.getLocalHost().getHostName()
					.toLowerCase();
			if (hostName.equals("wenzel-pc"))
				return "C:\\cygwin\\bin";
		} catch (UnknownHostException e) {
			fail(e.getMessage());
		}
		return null;
	}

	/**
	 * Read from a stream until it closes, and return the resulting string.
	 */
	private static String readerToString(BufferedReader reader) throws IOException {
		StringBuffer strBuf = new StringBuffer();
		char[] temp = new char[1024];
		int numRead = 0;
		while ((numRead = reader.read(temp)) != -1)
			strBuf.append(temp, 0, numRead);
		reader.close();
		return strBuf.toString();
	}

	private static String streamToString(InputStream is) throws IOException {
		return readerToString(new BufferedReader(new InputStreamReader(is)));
	}

	/**
	 * Compile a fragment of I9 source using the driver.
	 */
	public static File compile(String source, String[] params) {
		File sourceFile = null;

		try {
			sourceFile = File.createTempFile("test", ".i9");
			PrintWriter pw = new PrintWriter(new FileWriter(sourceFile));
			pw.println(source);
			pw.close();
			String absPath = sourceFile.getAbsolutePath();
			
			/* Construct a command line and run the driver */
			String[] args = new String[params.length + 9];

			for (int i=0; i<params.length; ++i)
				args[i] = params[i];

			args[args.length-9] = "-I";
			args[args.length-8] = getIotaIncludeDirectory();

			args[args.length-7] = "--arch";

			if (Configuration.ARCH == Architecture.I386)
				args[args.length-6] = "i386";
			else if (Configuration.ARCH == Architecture.X86_64)
				args[args.length-6] = "x86_64";
			else
				throw new InternalCompilerException("Unknown architecture!");		

			args[args.length-5] = "--target";

			if (Configuration.TARGET == Target.LINUX)
				args[args.length-4] = "linux";
			else if (Configuration.TARGET ==  Target.WINDOWS)
				args[args.length-4] = "windows";
			else if (Configuration.TARGET ==  Target.DARWIN)
				args[args.length-4] = "darwin";
			else
				throw new InternalCompilerException("Unknown target!");		
			
			
			args[args.length-3] = "--asmformat";

			if (Configuration.SYNTAX == AssemblySyntax.ATT_SYNTAX ||
				Configuration.TARGET == Target.DARWIN)
				args[args.length-2] = "att";
			else if (Configuration.SYNTAX == AssemblySyntax.INTEL_SYNTAX)
				args[args.length-2] = "intel";
			else
				throw new InternalCompilerException("Unknown assembly syntax!");

			args[args.length-1] = absPath;

			int result = Driver.run(args);
			if (result != 0)
				fail("Compilation failure!");

			return new File(absPath.substring(0, absPath.length()-3) + ".s");
		} catch (IOException e) {
			e.printStackTrace();
			fail("Unable to compile code fragment! - " + e.getMessage());
		} finally {
			if (sourceFile != null)
				sourceFile.delete();
		}
		return null;
	}
	
	/**
	 * Return the assembly fragment corresponding to a I9 source fragment.
	 * Throws away the first four lines
	 */
	public static String getAssembly(String source, String[] params) {
		File assemblyFile = compile(source, params); 

		try {
			BufferedReader br = new BufferedReader(new FileReader(assemblyFile));
			for (int i=0; i<4; ++i)
				br.readLine();
			return readerToString(br);
		} catch (IOException e) {
			e.printStackTrace();
			fail("Unable to compile code fragment! - " + e.getMessage());
			return null;
		} finally {
			if (assemblyFile != null)
				assemblyFile.delete();
		}		
	}

	/**
	 * Compile a fragment of I9 source using the driver,
	 * followed by a call to the assembler+linker.
	 */
	public File compileAndLink(String input) {
		return compileAndLink(new String[] { input });
	}

	public File compileAndLink(String[] input) {
		File[] assemblyFile = new File[input.length];
		for (int i=0; i<input.length; ++i)
			assemblyFile[i] = compile(input[i], !optimize ? new String[] { "-O" } : new String[] { });

		try {
			Runtime runtime = Runtime.getRuntime();
			File runtimeDirectory = new File(getIotaRuntimeDirectory());

			String asmPath = assemblyFile[0].getAbsolutePath();
			File binaryFile = new File(asmPath.substring(0, asmPath.length()-2));
			String cygwinPath = getCygwinPath();

			Process proc;

			if (cygwinPath == null) {
				String[] args = new String[3 + input.length];
				args[0] = "./linki9.sh";
				for (int i=0; i<input.length; ++i)
					args[i+1] = assemblyFile[i].getAbsolutePath();
				args[input.length+1] = "-o";
				args[input.length+2] = binaryFile.getAbsolutePath();
				proc = runtime.exec(args, null, runtimeDirectory);
			} else {
				String args = "export PATH=/bin; ./linki9.sh ";
				for (int i=0; i<input.length; ++i)
					args = args + assemblyFile[i].getAbsolutePath() + " ";
				args = args + "-o ";
				args = args + binaryFile.getAbsolutePath();

				args = args.replaceAll("\\\\", "/");
				args = args.replaceAll("C:/", "/cygdrive/c/");
				proc = runtime.exec(new String[] { cygwinPath + "\\bash.exe",
						"-c", args }, null, runtimeDirectory);
			}
			proc.waitFor();

			String compilerOutput = streamToString(proc.getInputStream());
			if (compilerOutput.trim().length() > 0)
				fail("Compilation failure : " + compilerOutput);

			return binaryFile;
		} catch (IOException e) {
			e.printStackTrace();
			fail("Unable to compile code fragment! - " + e.getMessage());
		} catch (InterruptedException e) {
			e.printStackTrace();
			fail("Unable to compile code fragment! - " + e.getMessage());
		} finally {
			if (assemblyFile != null) {
				for (int i=0; i<input.length; ++i)
					assemblyFile[i].delete();
			}
		}
		return null;
	}

	/**
	 * Execute a binary with the given input sequence and record the output
	 */
	public String runBinaryWithInput(File binaryFile, String input) {
		Runtime runtime = Runtime.getRuntime();

		try {
			Process proc = runtime.exec(new String[] { binaryFile.getAbsolutePath()
					+ (getCygwinPath() != null ? ".exe" : ""), input});
			OutputStream os = proc.getOutputStream();
			PrintWriter pw = new PrintWriter(new OutputStreamWriter(os));
			pw.print(input);
			pw.flush();
			pw.close();
			return streamToString(proc.getInputStream()) 
				+ streamToString(proc.getErrorStream());
		} catch (IOException e) {
			e.printStackTrace();
			fail("Unable to execute binary: " + e.getMessage());
		}

		return null;
	}

	/**
	 * Compile a fragment of i9 source, execute it with the given input sequence 
	 * and record the output
	 */
	public String compileLinkAndRun(String source, String input) {
		File binaryFile = compileAndLink(source);
		try {
			return runBinaryWithInput(binaryFile, input);
		} finally {
//			if (binaryFile != null)
	//			binaryFile.delete();
		}
	}

	/**
	 * Same as above, but with no input.
	 */
	public String compileLinkAndRun(String source) {
		return compileLinkAndRun(source, "");
	}

	public String executeFilewithInput (String filename, String input) {
		try {
			BufferedReader source = new BufferedReader(
					new FileReader(filename));
			String line = source.readLine();
			String s = "";
			while (line !=null) {
				s += line+"\n";
				line = source.readLine();
			}
			String output = compileLinkAndRun(s, input);
			return output;
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}
	
	@Test
	public void testHelloWorld() {
		String output = compileLinkAndRun(
				"use io;\n" +
				"main(args:int[][]) = {\n" +
				"   print(\"Hello world!\");\n" +
				"}"
		);
		assertEquals("Hello world!", output);
	}

	@Test
	public void testArrayOutOfBounds() {
		String output = compileLinkAndRun(
				"main(args:int[][]) = {\n" +
				"   i:(int, int) = (1,2);" +
				"   j:int = i 2;" +
				"}"
		);
		assertTrue(output.contains("Array index out of bounds"));
	}

	@Test
	public void testIO() {
		File binary = compileAndLink(
				"use io;\n" +
				"use conv;\n" +
				"main(args:int[][]) = {\n" +
				"	(i1:int, b1:bool) = parseInt(readln());\n" +
				"	(i2:int, b2:bool) = parseInt(readln());\n" +
				"   if (!b1 | !b2)\n" +
				"       print(\"Could not parse input!\");\n" +
				"   else\n" +
				"       print(unparseInt(i1 + i2));\n" +
				"}"
		);
		try {
			assertEquals("21", 
					runBinaryWithInput(binary, "16\n5\n"));
			assertEquals("Could not parse input!", 
					runBinaryWithInput(binary, "16\nb\n"));
		} finally {
			binary.delete();
		}
	}
	
	@Test
	public void testBellmanFord() {
		File binary = compileAndLink( 
			"use io;" +
			"BellmanFord(predecessor:int[], weight:int[][], edges:(int, int)[],  source:int):int[] = {" +
			"	n:int = 5;" +
			"	m:int = 6;" +
			"	distance:int[n];" +
			"" +
			"" +
			"	j:int=0;" +
			"	while(j < n) {" +
			"		if(j == source)" +
			"			distance j = 0;" +
			"		else" +
			"			distance j = 10000;" +
			"		predecessor(j) = -1;" +
			"		j = j+1;" +
			"	}" +
			"" +
			"	i:int = 0;" +
			"	j = 0;" +
			"	while(i<= n-2) {" +
			"		while(j<m) {" +
			"			u:int = edges(j) 0;" +
			"			v:int = edges(j) 1;" +
			"			if(distance(u) + (weight u v) < distance(v)) {" +
			"				distance(v) = distance(u) + (weight u v);" +
			"				predecessor(v) = u;" +
			"			}" +
			"			j = j+1;" +
			"		}" +
			"		i = i+1;" +	
			"	}" +
			"	" +
			"" +
			"	i = 0;" +
			"	while(i < m) { " +
			"		u:int = edges(i) 0;" +
			"		v:int = edges(i) 1;" +
			"		if(distance(u) + (weight u v) < distance(v))" +
			"			print(\"Error: graph contains a negative-weight cycle\");" +
			"		i = i+1;" +
			"	}" +
			"	return distance;\n" +
			"}" +
			"" +
			"execute():int[] = {" +
			"	weight:int[][] = (" +
			"		(0, 7, 10000, 10000, 1)," +
			"		(7, 0, 1, 10000, 8)," +
			"		(10000, 1, 0, 2, 10000)," +
			"		(10000, 10000, 2, 0, 2)," +
			"		(1, 8, 10000, 2, 0)" +
			"	);" +
			"	predecessor:int[length weight];" +
			"	source:int = 0;" +
			"	edges:(int,int)[] = (" +
			"		(0, 1)," +
			"		(0, 4)," +
			"		(1, 2)," +
			"		(1, 4)," +
			"		(2, 3)," +
			"		(3, 4)" +
			"	);" +
			"	return BellmanFord(predecessor, weight, edges, source);" +
			"}" +
			"main(args:int[][]) = {" +
			"	a:int[]=execute();\n   " +
			"	b:int[]=(0,7,8,10,1);\n " +
			"	i:int = 0;\n" +
			"	r:bool = true;" +
			"	blen:int = length b;\n" +
			"	c:int = 123;" +
			"	while(i < blen) {" +
			"		if(a i != b i) {" +
			"			r = false;\n" +
			"		}  \n" +
			"		i=i+1;\n" +
			"	}" +
			"	if(r)" +
			"		print(\"correct\");\n" +
			"}");
		try {
			assertEquals("correct", runBinaryWithInput(binary, ""));
		} finally {
			binary.delete();
		}
	}
	
	@Test
	public void testCallwithArray() {
		File binary = compileAndLink( 
				"use io;" +
				"a(x:int[]) = {" +
				"  	i:int = x 1;" +
				"	j:int = x 3;" +
				"	k:int = x 4;" +
				"	print j;" +
				"}" +
				"main(args:int[][])={" +
				"	y:int[]=('1','2','3','4','5');" +
				"	a(y);" +
				"}");
		try {
			assertEquals("4", runBinaryWithInput(binary, ""));
		} finally {
			binary.delete();
		}
	}
	
	
	@Test
	public void testPrim() {
		// O(n^2) version of Prim's algorithm

		File binary = compileAndLink( 
			"use io;" +
			"use conv;" +
			"" +
			"updateDistances(adj:int[][], d:int[], treeSucc:int[], target:int) = {" +
			"    n:int = length(adj);" +
			"    i:int = 0;" +
			"    while (i<n) {" +
			"        if (adj target i > 0 &" +
			"            d i > adj target i) {" +
			"            d i = adj target i;" +
			"            treeSucc i = target;" +
			"        }" +
			"        i = i+1;" +
			"    }" +
			"}" +
			"" +
			"prim(adj:int[][]) : int = {" +
			"    (cost:int, n:int) = (0, length(adj));" +
			"    contained:bool[n];" +
			"    treeSucc:int[n];" +
			"" +
			"    d:int[n];" +
			"    large_val:int = 100000;" +
			"" +
			"    i:int = 0;" +
			"    while (i < n) {" +
			"        d(i) = large_val;" +
			"        contained(i) = false;" +
			"        i = i+1;" +
			"    }" +
			"" +
			"    contained(0) = true;" +
			"    updateDistances(adj, d, treeSucc, 0);" +
			"" +
			"    i = 0;" +
			"    while (i<n-1) {" +
			"        (j:int, min:int) = (0, -1);" +
			"        while (j<n) {" +
			"            if (!contained j)" +
			"                if (min == -1 | d(j) < d(min))" +
			"                    min = j;" +
			"            j = j+1;" +
			"        }" +
			"" +
			"        contained(min) = true;" +
			"        updateDistances(adj, d, treeSucc, min);" +
			"        cost = cost + d min;" +
			"" +
			"        i = i+1;" +
			"    }" +
			"" +
			"" +
			"    return cost;" +
			"}" +
			"" +
			"main(args:int[][]) = {" +
			"    adj:int[7][7];" +
			"    i:int = 0;" +
			"    while (i<7) {" +
			"        j:int = 0;" +
			"        while (j<7) {" +
			"            adj i j = 0;" +
			"            j = j+1;" +
			"        }" +
			"        i = i+1;" +
			"    }" +
			"    adj 0 1 = 7;  adj 1 0 = 7;" +
			"    adj 0 3 = 5;  adj 3 0 = 5;" +
			"    adj 1 2 = 8;  adj 2 1 = 8;" +
			"    adj 1 3 = 9;  adj 3 1 = 9;" +
			"    adj 1 4 = 7;  adj 4 1 = 7;" +
			"    adj 2 4 = 5;  adj 4 2 = 5;" +
			"    adj 3 4 = 15; adj 4 3 = 15;" +
			"    adj 3 5 = 6;  adj 5 3 = 6;" +
			"    adj 4 5 = 8;  adj 5 4 = 8;" +
			"    adj 4 6 = 9;  adj 6 4 = 9;" +
			"    adj 5 6 = 11; adj 6 5 = 11;" +
			"    print(unparseInt(prim adj));" +
			"}");
		try {
			assertEquals("39", runBinaryWithInput(binary, ""));
		} finally {
			binary.delete();
		}
	}

	@Test
	public void testItoa() {
		/// Check if IMOD/IDIV are handled correctly in the allocator
		File binary = compileAndLink(
			"use io;" +
			"itoa(i:int) : int[] = {" +
			"    string:int[0];" +
			"" +
			"    while (true) {" +
			"        newString:int[length string + 1];" +
			"        j:int = 0;" +
			"        while (j<length string) {" +
			"            newString(j+1) = string j;" +
			"            j = j+1;" +
			"        }" +
			"        string = newString;" +
			"        string 0 = '0' + (i % 10);" +
			"        i = i / 10;" +
			"" +
			"        if (i == 0) " +
			"            break;" +
			"    }" + 
			"    return string;" + 
			"}" +
			"main(args:int[][]) = {" +
			"    print(itoa(153));" +
			"}");
		try {
			assertEquals("153", runBinaryWithInput(binary, ""));
		} finally {
			binary.delete();
		}
	}

	@Test
	public void testUnderscore() {
		File binary = compileAndLink(
		"use io;" +
		"use conv;" +
		"a(i:int, _, (j:int, _)):int = {"+
		"	print(unparseInt(i+j));"+
		"	return i;"+
		"	print(unparseInt(i*j));"+
		"}"+
		"main(args:int[][]) = {"+
		"	_ = a(5, 2, (9, 4));"+
		"}");
		try {
			assertEquals("14", runBinaryWithInput(binary, ""));
		} finally {
			binary.delete();
		}
	}

	@Test
	public void testManyArguments() {
		File binary = compileAndLink(
		"use io;" +
		"use conv;" +
		"a(p1:int, p2:int, p3: int, p4:int, p5:int, p6:int, p7:int, p8:int, p9:int): (int, int, int) = {"+
		"	print(unparseInt(p1));"+
		"	print(unparseInt(p2));"+
		"	print(unparseInt(p3));"+
		"	print(unparseInt(p4));"+
		"	print(unparseInt(p5));"+
		"	print(unparseInt(p6));"+
		"   return (p7, p8, p9);" +
		"}"+
		"main(args:int[][]) = {"+
		"	(i:int, j:int, k:int) = a(1,2,3,4,5,6,7,8,9);"+
		"	print(unparseInt(i));"+
		"	print(unparseInt(j));"+
		"	print(unparseInt(k));"+
		"}");
		try {
			assertEquals("123456789", runBinaryWithInput(binary, ""));
		} finally {
			binary.delete();
		}
	}	
	
	@Test
	public void testTupleIndexing() {
		// Mostly tests the tuple indexing feature
		File binary = compileAndLink(
				"use assert;" + 
				"" + 
				"main(args:int[][]) = {" + 
				"	a:(int, bool)[] = (" + 
				"		((1, true), (5, false))," + 
				"		((23, false), (1, true), (6, false))" + 
				"	) 1;" + 
				"" + 
				"	(i:int, _, j:int, k:bool) = (0, (40, true), 0, false);" + 
				"" + 
				"	while (i<3) {" + 
				"		(l:int, m:bool) = a i;" + 
				"		j = j+l;" + 
				"		k = k | m;" + 
				"		i = i+1;" + 
				"	}" + 
				"" + 
				"	assert(j == 30);" + 
				"	assert(k | (j==29) == true);" + 
				"}"); 
		try {
			assertEquals("", runBinaryWithInput(binary, ""));
		} finally {
			binary.delete();
		}
	}
	
	@Test
	public void testTupleAssignment () {
		File binary = compileAndLink(
				"use io;" +
				"a(x:int[]) = {" +
				"  	(x, x 2) = (x, x 1);" +
				"	y:(int, int, int) = ('D','E','F');" +
				"	z:(int, int, int) = ('H','I','J');" +
				"	(y,z) = (z,y);" +
				"	print (x 0);" +
				"	print (x 1);" +
				"	print (x 2);" +
				"	print (y 0);" +
				"	print (y 1);" +
				"	print (y 2);" +
				"}" +
				"main(args:int[][]) = {" +
				"	a(('A','B','C'));" +
				"}");
		try {
			assertEquals("ABBHIJ", runBinaryWithInput(binary, ""));
		} finally {
			binary.delete();
		}
	}
	
	@Test
	public void testNegDivision() {
		// Check that SAR is used for the strength-reduced division
		// (instead of SHR) and that the sign of EDX is initialized properly
		File binary = compileAndLink(
			"use assert;"+
			"main(args:int[][]) = {"+
			"    (a:int, b:int) = (0, 8);"+
			"    (c:int, d:int) = (-8, a-b);"+
			"    assert((c / 8) == -1);"+
			"    assert((d / 8) == -1);"+
			"}");
		try {
			assertEquals("", runBinaryWithInput(binary, ""));
		} finally {
			binary.delete();
		}
	}

	@Test
	public void testStringTable() {
		File binary = compileAndLink(
				"use assert;"+
				"a() : int[] = {"+
				"   (i:int, b:int[]) = (0, \"Hello\");"+
				"   c:int[length b];"+
				"   // Make a copy\n"+
				"   while (i<length b) {"+
				"       c i = b i;"+
				"       i = i+1;"+
				"   }"+
				"   // Modify B and return the copy\n"+
				"   b 0 = 'B';"+
				"   assert(b 0 == 'B');"+
				"   return c;"+
				"}"+
				"main(args:int[][]) = {"+
				"   (_, b:int[]) = (a(), a()); // Ignore 1st eval\n"+
				"   assert(b 0 == 'H');"+
				"}");
		try {
			assertEquals("", runBinaryWithInput(binary, ""));
		} finally {
			binary.delete();
		}
	}

	@Test
	public void testReturnArray() {
		File binary = compileAndLink(
				"use io;" +
				"a():int[] = {" +
				"  	y:int[] = ('1','2','3','4','5');" +
				"	return y;" +
				"}" +
				"main(args:int[][]) = {" +
				"	array:int[] = a();" +
				"	print array;" +
				"}"	);
		try {
			assertEquals("12345", runBinaryWithInput(binary, ""));
		} finally {
			binary.delete();
		}
	}

	@Test
	public void testArrayReassign() {
		File binary = compileAndLink(
				"use assert;" +
				"use io;" + 
				"" + 
				"main(args:int[][]) = {" + 
				"	a:(int[], int)[][];" + 
				"	b:(int[], int)[] = (\n" +
				"		(\"FOO\", 1)," + 
				"		(\"BAR\", 1)," + 
				"		(\"BAZ\", 1));" + 
				"	c:(int[], int)[] = (\n" +
				"		(\"GRAULT\", 0)," + 
				"		(\"GARPLY\", 0));" +
				"	a = loopy(b, c);\n" +
				"	print(a 0 0 0);" +
				"}\n" +
				"loopy(b:(int[], int)[], c:(int[], int)[]):(int[],int)[][] = {" +
				"	a:(int[], int)[][] = (b, c);" +
				"	i:int;" +
				"	j:int;" +
				"	while(length a < 3) {" +
				"		i=0;" +
				"		while(i < length a) {" +
				"			j=0;" +
				"			while(j < length a i) {" +
				"				if( j % 2 == 0)" +
				"					a i j = (a i j 0, 1);" +
				"				j = j + 1;" +
				"			}" +
				"			i = i+1;" +
				"		}" +
				"		if(length a == 2)" +
				"			a = (c, b,c);" +
				"		else if (length a == 3)" +
				"			a = (b, c, c, b);" +
				"		else" +
				"			return (c, c, b, c, b);" +
				"	}" + 
				"	return (c, c, b, c); \n"+
				"}"); 
		try {
			assertEquals("GRAULT", runBinaryWithInput(binary, ""));
		} finally {
			binary.delete();
		}
	}

	@Test
	public void testReturnArrayQuickSort() {
		File binary = compileAndLink(
				"//Quicksort Implementation\n" +
				"use io;" +
				"use conv;" +
				"printArray(arr:int[]) = {" +
				"	i:int = 0;" +
				"	while(i<length arr) {" +
				"		print(unparseInt(arr i));" +
				"		print(\" \");" +
				"		i = i+1;" +
				"	}" +
				"	println();" +
				"}" +
				"" +
				"Quicksort(A:int[], p:int, r:int) = {" +
				"	if(p<r) {" +
				"		q:int = Partition(A, p, r);" +
				"		Quicksort(A, p, q-1);" +
				"		Quicksort(A, q+1, r);" +
				"	}" +
				"}" +
				"" +
				"Partition(A:int[], p:int, r:int):int = {" +
				"	x:int = A r;" +
				"	i:int = p-1;" +
				"	j:int = p;" +
				"	while(j<=r-1) {" +
				"		if(A j <= x) {" +
				"			i = i+1;" +
				"			aux:int = A i;" +
				"			A i = A j;" +
				"			A j = aux;" +
				"		}" +
				"		j = j+1;" +
				"	}" +
				"	aux:int = A(i+1);" +
				"	A(i+1) = A r;" +
				"	A r = aux;" +
				"	return i+1;" +
				"}" +
				"" +
				"main(args:int[][]) = {" +
				"	A:int[] = (2,8,7,1,3,5,6,4);" +
				"	Quicksort(A, 0, length A - 1);" +
				"	printArray(A);" +
				"}");
		try {
			assertEquals("1 2 3 4 5 6 7 8", runBinaryWithInput(binary, "").trim());
		} finally {
			binary.delete();
		}
	}

	@Test
	public void testReturnArrayDivision() {
		File binary = compileAndLink(
				"use io;" +
				"use conv;" +
				"" +
				"div(divisor:int, dividend:int, decimals:int, rez:int[], size:int):(int[], int) = {" +
				"	if(decimals == 0 |  divisor == 0)" +
				"		return (rez, size);" +
				"	q:int = divisor/dividend;" +
				"	rez size = q;" +
				"	size = size + 1;" +
				"	remainder:int = divisor - q*dividend;" +
				"	return div(10 * remainder, dividend, decimals-1, rez, size);" +
				"}" +
				"" +
				"main(args:int[][]) = {" +
				"	(divisor:int, flag1:bool) = parseInt(readln());" +
				"	(dividend:int, flag2:bool) = parseInt(readln());" +
				"	(decimals:int, flag3:bool) = parseInt(readln());" +
				"   if (!flag1 | !flag2 | !flag3)\n" +
				"       println(\"Could not parse input!\");\n" +
				"   else {" +
				"		i:int = 1;" +
				"		auxRez:int[decimals+1];" +
				"		(rez:int[decimals+1], size:int) = div(divisor, dividend, decimals, auxRez, 0);" +
				"		print(unparseInt(rez 0));" +
				"		print(\".\");" +
				"" +
				"		while(i<decimals) {" +
				"			print(unparseInt(rez i));" +
				"			i = i + 1;" +
				"		}" +
				"	}" +
				"}"	);
		try {
			assertEquals("2.345", runBinaryWithInput(binary, "2345\n1000\n4"));
		} finally {
			binary.delete();
		}
	}

	@Test
	public void testReturnArrayOfArraysDynamicProg() {
		File binary = compileAndLink(
				"//Dynamic programming\n" +
				"use io;" +
				"use conv;" +
				"" +
				"Subset_Sum(weights:int[], n:int, W:int):int[][] = {" +
				"	OPT:int[n+1][W + 1];" +
				"	M:int[n+1][W + 1];" +
				"	w:int = 0;" +
				"	while(w<=W) {" +
				"		M 0 w = 0;" +
				"		w = w + 1;" +
				"	}" +
				"	i:int = 1;" +
				"	while(i<=n) {" +
				"		w = 0;" +
				"		while(w<=W) {" +
				"			if(w < weights i)" +
				"				M i w = M (i-1) w;" +
				"" +
				"			else" +
				"				M i w = max(M (i-1) w, (weights i) + M (i-1) (w - (weights i)));" +
				"			w = w+1;" +
				"		" +
				"		}" +
				"		i = i+1;" +
				"	}" +
				"	return M;" +
				"}" +
				"" +
				"max(a:int, b:int):int = {" +
				"	if(a>b)" +
				"		return a;" +
				"	return b;" +
				"}" +
				"main(args:int[][]) = {" +
				"	weights:int[] = (0, 2, 2, 3);" +
				"	W:int = 6;" +
				"	n:int = 3;" +
				"	OPT:int[][] = Subset_Sum(weights, n, W);" +
				"	i:int = n;" +
				"	w:int = 0;" +
				"	" +
				"	while(w<=W) {" +
				"		print(\" \");" +
				"		print(unparseInt(OPT i w));" +
				"		print(\" \");" +
				"		w = w+1;" +
				"		" +
				"	}" +
				"}"	);
		try {
			assertEquals(" 0  0  2  3  4  5  5 ", runBinaryWithInput(binary, ""));
		} finally {
			binary.delete();
		}
	}

	@Test
	public void testMatrixMultiplication() {
		File binary = compileAndLink(
				"use io;" +
				"use conv;" +
				"" +
				"multiply(a:int[][], b:int[][]):int[][] = {" +
				"	mA:int = length a;" +
				"	nA:int = length a 0;" +
				"	mB:int = length b;" +
				"" +
				"	nB:int = length b 0;" +
				"	if(nA != mB)" +
				"" +
				"		println(\"Cannot multiply matrices! # of columns of A not equal to # of rows of B\");" +
				"	else {" +
				"		c:int[mA][nB];" +
				"		i:int = 0;" +
				"		j:int;" +
				"		r:int;" +
				"		while(i < mA) {" +
				"			j = 0;" +
				"			while(j < nB) {" +
				"				c i j = 0;" +
				"				r = 0;" +
				"				while(r < nA) {" +
				"					c i j = (c i j) + (a i r)*(b r j);" +
				"					r = r+1;" +
				"				} " +
				"				j = j+1;" +
				"			}" +
				"			i = i+1;" +
				"		}" +
				"		return c;" +
				"	}" +
				"}" +
				"" +
				"printMatrix(a:int[][]) = {" +
				"	m:int = length a;" +
				"	n:int = length a 0;" +
				"	i:int = 0;" +
				"	j:int = 0;" +
				"	while(i<m) {" +
				"		j = 0;" +
				"		while(j<n) {" +
				"			print(unparseInt(a i j));print(\" \");" +
				"			j = j+1;" +
				"		}" +
				"		println(\"\");" +
				"		i = i+1;" +
				"	}" +
				"}" +
				"" +
				"main(args:int[][]) = {	M:int[][] = (" +
				"		(1, 2, 3, 4, 5)," +
				"		(6, 7, 8, 9, 10)," +
				"		(11, 12, 13, 14, 15)," +
				"		(16, 17, 18, 19, 20)," +
				"		(21, 22, 23, 24, 25)" +
				"	);" +
				"	result:int[][] = multiply(M, M);" +
				"	printMatrix(result);" +
				"}"	);
		try {
			if (getCygwinPath()==null) {
				assertEquals("215 230 245 260 275 \n" +
					"490 530 570 610 650 \n" +
					"765 830 895 960 1025 \n" +
					"1040 1130 1220 1310 1400 \n" +
					"1315 1430 1545 1660 1775 \n", runBinaryWithInput(binary, ""));
			}
			else {
				assertEquals("215 230 245 260 275 \r\n" +
						"490 530 570 610 650 \r\n" +
						"765 830 895 960 1025 \r\n" +
						"1040 1130 1220 1310 1400 \r\n" +
						"1315 1430 1545 1660 1775 \r\n", runBinaryWithInput(binary, ""));
			}
		} finally {
			binary.delete();
		}
	}
	
	@Test
	public void testDeterminant() {
		File binary = compileAndLink(
				"use io;" +
				"use conv;" +
				"" +
				"subMatrix(A:int[][], removeLine:int):int[][] = {" +
				"	m:int = length A;" +
				"	n:int = length A 0;" +
				"	i:int = 0;" +
				"	j:int;" +
				"	result:int[m-1][n-1];" +
				"	while(i<n) {" +
				"		j = 1;" +
				"		if(i!=removeLine) {" +
				"			while(j<m) {" +
				"				if(i > removeLine)" +
				"					(result (i-1) (j-1)) = (A i j);" +
				"				else" +
				"					result i (j-1) = A i j;" +
				"				j = j+1;" +
				"			}" +
				"		}" +
				"		i = i+1;" +
				"	}	" +
				"return result;" +
				"" +
				"}" +
				"" +
				"det(A:int[][]):int = {" +
				"	m:int = length A;" +
				"	n:int = length A 0;" +
				"	if (m!=n){" +
				"		println(\"Error! Cannot compute determinant of non-square matrix\");" +
				"		return -1;	" +
				"}" +
				"	if(m == 1) {" +
				"		return A 0 0;" +
				"	}" +
				"	i:int = 0;" +
				"	j:int = 0;" +
				"	sum:int = 0;" +
				"	sign:int = -1;" +
				"	while(i<m) {" +
				"		sign = -sign;" +
				"		sum = sum + sign * (A i 0) * det(subMatrix(A, i));" +
				"		i = i + 1;" +
				"	}" +
				"	return sum;" +
				"}" +
				"" +
				"main(args:int[][]) = {" +
				"	M:int[][] = (" +
				"		(1, 2, 3, 4, 5)," +
				"		(6, 8, 10, 12, 15)," +
				"		(19, 24, 30, 36, 43)," +
				"		(51, 60, 70, 81, 92)," +
				"		(104, 127, 141, 155, 170)" +
				"	);" +
				"	determ:int = det(M);" +
				"	print(unparseInt(determ));" +
				"}");
		try {
			assertEquals("167", runBinaryWithInput(binary, ""));
		} finally {
			binary.delete();
		}
	}
	
	@Test
	public void testStaffTestcase1() {
		/// Staff test case we failed in PA5
		File binary = compileAndLink(
			"use io;"+
			"use conv;"+
			"use assert;"+
			""+
			"test(msg: int[], result: bool) = {"+
			"    print(\"Testing:\");"+
			"    println(msg);"+
			"    assert(result);"+
			"}"+
			""+
			"main(args: int[][]) = {"+
			"    a:int = 30;"+
			"    b:int = 42;"+
			"    test(\"<\", a < b == true);"+
			"    test(\"<=\", a <= b == true);"+
			"    test(\">\", a > b == false);"+
			"    test(\">=\", a >= b == false);"+
			"    test(\"==\", !(a == b));"+
			"    test(\"!=\", a != b);"+
			"}"
		);
		try {
			if (getCygwinPath()==null) {
				assertEquals("Testing:<\nTesting:<=\nTesting:>\n"+
					"Testing:>=\nTesting:==\nTesting:!=\n",
					runBinaryWithInput(binary, ""));
			}
			else {
				assertEquals("Testing:<\r\nTesting:<=\r\nTesting:>\r\n"+
						"Testing:>=\r\nTesting:==\r\nTesting:!=\r\n",
						runBinaryWithInput(binary, ""));
			}
		} finally {
			binary.delete();
		}
	}

	@Test
	public void testStaffTestcase2() {
		/// Staff test case we failed in PA5
		File binary = compileAndLink(
			"use io;"+
			"unreachable() : bool = {"+
			"	println(\"Should never get here!\");"+
			"	return true;"+
			"}"+
			"reachable() = {" +
			"	print(\"This is OK\");" +
			"}" +
			"main(args: int[][]) = {"+
			"	f:bool = false;"+
			"	if (!f | !unreachable())"+
			"		reachable();"+
			"}"
		);
		try {
			assertEquals("This is OK",
					runBinaryWithInput(binary, ""));
		} finally {
			binary.delete();
		}
	}

	@Test
	public void testPermutations() {
		File binary = compileAndLink(
				"use io;" +
				"use conv;" +
				"" +
				"range(start:int, end:int):int[] = {" +
				"	i:int = 0;" +
				"	result:int[end-start];" +
				"	while(i < end-start) {" +
				"		result(i) = i + start;" +
				"		i = i+1;" +
				"" +
				"	}" +
				"	return result;" +
				"}" +
				"" +
				"Fannkuch(n:int , ii:int, okPrint:bool):int = {" +
				"	check:int = 0;" +
				"   perm:int[];" +
				"   perm1:int[];" +
				"   count:int[n+1];" +
				"	maxFlipsCount:int = 0;" +
				"   m:int = n;" +
				"" +
				"	perm1 = range(0, n);" +
				"	r:int = n;" +
				"	if (!okPrint) {" +
				"		perm1 ii = n;" +
				"		perm1(n) = ii;" +
				"	}" +
				"   " +
				"   while (true) {" +
				"	// write-out the first 30 permutations\n" +
				"		if (okPrint){" +
				"			aux:int = 0;" +
				"			check = 0;" +
				"			while(aux < length perm1) {" +
				"				print(unparseInt(perm1(aux) + 1));" +
				"				aux = aux +1;" +
				"				check = check+1;" +
				"				if(check > 30)" +
				"					break;" +
				"			}" +
				"			print(\"\\n\");" +
				"      	}" +
				"" +
				"      	while (r != 1){" +
				"      		count(r-1) = r;" +
				"      		r = r-1;" +
				"      	}" +
				"" +
				"		if(m<length perm1)" +
				"		if ((perm1(0) != 0) & (perm1(m) != m)){" +
				"			perm = perm1;" +
				"			flipsCount:int = 0;" +
				"			k:int = perm(0);" +
				"			while(true){" +
				"				i:int = 1;" +
				"            	j:int = k - 1;" +
				"				while (i < j) {" +
				"					tmp:int = perm(i);" +
				"                	perm(i) = perm(j);" +
				"                	i = i + 1;" +
				"                	perm(j) = tmp;" +
				"                	j = j - 1;" +
				"				}" +
				"				flipsCount = flipsCount + 1;" +
				"            	tmp:int = perm k;" +
				"            	perm k = k;" +
				"				k = tmp;" +
				"				if(k == 0)" +
				"	            	break;" +
				"	        }" +
				"	        if (flipsCount > maxFlipsCount) {" +
				"            	maxFlipsCount = flipsCount;" +
				"           }" +
				"        }" +
				"" +
				"		while(true) {" +
				"			if (r >= m) {" +
				"				return maxFlipsCount;" +
				"			}" +
				"			perm0:int = perm1 0;" +
				"			i:int = 0;" +
				"         	while (i < r) {" +
				"            	perm1 i = perm1 (i+1);" +
				"            	i = i + 1;" +
				"         	}" +
				"         	perm1(r) = perm0;" +
				"         	count r = (count r) - 1;" +
				"         	if (count(r) > 0)" +
				"         		break;" +
				"         	r = r+1;" +
				"       }" +
				"   }" +
				"}" +
				"main(args:int[][]) = {" +
				"	n:int = 3;" +
				"	result:int = Fannkuch(n,0,true);" +
				"	print(result);" +
				"}"	);
		try {
			if (getCygwinPath()==null) 
				assertEquals("123\n" +
					"213\n" +
					"231\n" +
					"321\n" +
					"312\n" +
					"132", runBinaryWithInput(binary, "").trim());
			else
				assertEquals("123\r\n" +
						"213\r\n" +
						"231\r\n" +
						"321\r\n" +
						"312\r\n" +
						"132", runBinaryWithInput(binary, "").trim());
		} finally {
			binary.delete();
		}
	}

	@Test
	public void testTravelingSalesman() {
		File binary = compileAndLink(
				"use io;" +
				"use conv;" +
				"range(start:int, end:int):int[] = {" +
				"	i:int = 0;" +
				"	result:int[end-start];" +
				"	while(i < end-start) {" +
				"		result(i) = i + start;" +
				"		i = i+1;	" +
				"	}" +
				"	return result;" +
				"}" +
				"" +
				"traveling_salesman(n:int , ii:int, okPrint:bool, distance:int[][]):int = {" +
				"	check:int = 0;" +
				"	perm:int[];" +
				"	perm1:int[];" +
				"	count:int[n+1];" +
				"	maxFlipsCount:int = 0;" +
				"  	m:int = n;" +
				"	perm1 = range(0, n);" +
				"  	r:int = n;" +
				"	if (!okPrint) {" +
				"   		perm1 ii = n;" +
				"		perm1(n) = ii;" +
				"	}" +
				"   min:int = 2147483647;" +
				"   while (true) {" +
				"   		dist:int = 0;" +
				"		if (okPrint){" +
				"			aux:int = 0;" +
				"			check = 0;" +
				"			prev:int = perm1 0;" +
				"			while(aux < length perm1) {" +
				"				dist = dist + distance prev (perm1 aux);" +
				"				prev = perm1 aux;" +
				"				aux = aux +1;" +
				"				check = check+1;" +
				"				if(check > 30)" +
				"					break;" +
				"			}" +
				"      	}" +
				"      	if(min>dist)" +
				"      		min = dist;" +
				"      	while (r != 1){" +
				"      		count(r-1) = r;" +
				"      		r = r-1;" +
				"      	}" +
				"      	if(m<length perm1)" +
				"      	if ((perm1(0) != 0) & (perm1(m) != m)){" +
				"      		perm = perm1;" +
				"         	flipsCount:int = 0;" +
				"         	k:int = perm(0);" +
				"         	" +
				"         	while(true){" +
				"				i:int = 1;" +
				"            	j:int = k - 1;" +
				"                while (i < j) {" +
				"                	tmp:int = perm(i);" +
				"                	perm(i) = perm(j);" +
				"                	i = i + 1;" +
				"                	perm(j) = tmp;" +
				"                	j = j - 1;" +
				"                }" +
				"                " +
				"                flipsCount = flipsCount + 1;" +
				"            	tmp:int = perm k;" +
				"            	perm k = k;" +
				"            	k = tmp;" +
				"            	if(k == 0)" +
				"	            	break;" +
				"	        }" +
				"	        " +
				"	        if (flipsCount > maxFlipsCount) {" +
				"            	maxFlipsCount = flipsCount;" +
				"            }" +
				"        }" +
				"" +
				"		while(true) {" +
				"			if (r >= m)" +
				"				return  min;" +
				"			perm0:int = perm1 0;" +
				"			i:int = 0;" +
				"         	while (i < r) {" +
				"            	perm1 i = perm1 (i+1);" +
				"            	i = i + 1;" +
				"         	}" +
				"         	perm1(r) = perm0;" +
				"         	" +
				"         	count r = (count r) - 1;" +
				"         	if (count(r) > 0)" +
				"         		break;" +
				"         	r = r+1;" +
				"       }" +
				"   }" +
				"" +
				"}" +
				"" +
				"main(args:int[][]) = {" +
				"	distance:int[][] = (" +
				"		( 0, 3, 5,48,48, 8, 8, 5, 5, 3)," +
				"		( 3, 0, 3,48,48, 8, 8, 5, 5, 0)," +
				"		( 5, 3, 0,72,72,48,48,24,24, 3)," +
				"		(48,48,74, 0, 0, 6, 6,12,12,48)," +
				"		(48,48,74, 0, 0, 6, 6,12,12,48)," +
				"		( 8, 8,50, 6, 6, 0, 0, 8, 8, 8)," +
				"		( 8, 8,50, 6, 6, 0, 0, 8, 8, 8)," +
				"		( 5, 5,26,12,12, 8, 8, 0, 0, 5)," +
				"		( 5, 5,26,12,12, 8, 8, 0, 0, 5)," +
				"		( 3, 0, 3,48,48, 8, 8, 5, 5, 0)" +
				"	);" +
				"	n:int = length distance;" +
				"	min:int = traveling_salesman(n,0,true,distance);" +
				"	print(unparseInt(min));" +
				"}");
		try {
			assertEquals("25", runBinaryWithInput(binary, ""));
		} finally {
			binary.delete();
		}
	}
	
	@Test
	public void testGaussSeidel() {
		File binary = compileAndLink(
			"use io;"+
			"use conv;"+
			"// 200 iterations of an integer \"Gauss-Seidel solver\"\n"+
			"// for a 2D Poisson equation on a 1024^2 grid.\n"+
			"main(args:int[][]) = {"+
			"	dim : int = 1024;"+
			"	data:int[dim*dim];"+
			"	(i:int, j:int) = (0, 0);"+
			"	while (i<dim*dim) {"+
			"		data i = 0;"+
			"		i = i+1;"+
			"	}"+
			"	k:int = 0;"+
			"	while (k < 200) {"+
			"		// Set boundary conditions\n"+
			"		data(dim/2*dim + dim/2) = 1000;"+
			"		i = 0;"+
			"		while (i < dim) {"+
			"			data i = 0;"+
			"			data (i*dim) = 0;"+
			"			data (i*dim + dim-1) = 0;"+
			"			data (dim*(dim-1) + i) = 0;"+
			"			i = i+1;"+
			"		}"+
			"		/// Gauss-seidel step\n"+
			"		i = 1;"+
			"		while (i < dim-1) {"+
			"			j = 1;"+
			"			while (j < dim-1) {"+
			"				data(i*dim + j) ="+
			"					(data(i*dim+j + 1)"+
			"					+ data(i*dim+j - 1)"+
			"					+ data(i*dim+j + dim)"+
			"					+ data(i*dim+j - dim)) / 4;"+
			"				j = j+1;"+
			"			}"+
			"			i = i+1;"+
			"		}"+
			"		k = k+1;"+
			"	}"+
			"   sum:int = 0;"+
			"   // Print the sum of all elements\n"+
			"   i = 1;"+
		    "   while (i < dim*dim) {"+
		    "       sum = sum + data i;"+
		           "i = i+1;"+
		    "   }"+
		    "   print(unparseInt(sum));"+
			"}");
		try {
			assertEquals("26354", runBinaryWithInput(binary, ""));
		} finally {
			binary.delete();
		}
	}

	@Test
	public void testGraphColoring() {
		File binary = compileAndLink(
				"use io; "+
				"use conv; "+
				" "+
				"append(array:int[], x:int):int[] = { "+
				"    result:int[(length array) + 1]; "+
				"    i:int = 0; "+
				"    while(i < length array) { "+
				"        result i = array i; "+
				"        i = i+1; "+"    } "+
				"    result (length array) = x; "+	
				"    return result; "+
				"} "+
				" "+
				"print_coloring(edges:(int, int)[],start:int[], n:int, k:int):int = { "+
				"    if (n <= 0){ "+
				"        i:int = 0; "+
				"        flag:bool = true; "+
				"        while(i < length edges) { "+
				"            if(start((edges i 0)-1) == start((edges(i) 1)-1)) "+
				"                flag = false; "+
				"            i = i+1; "+
				"        } "+
				"        if (flag){ "+
				"            return 1; "+
				"        } "+
				"        return 0; "+
				"    } "+
				"    else { "+
				"        i:int = 1; "+
				"        number_of_solns:int = 0; "+
				"        while(i <= k) { "+
				"            newStart:int[] = append(start, i); "+
				"            number_of_solns = number_of_solns + print_coloring(edges, newStart, n - 1, k); "+
				"            i = i+1; "+
				"        } "+
				"        return number_of_solns; "+
				"    } "+
				"} "+
				" "+
				"main(args:int[][]) = { "+
				"    start:int[] = (); "+
				"    edges:(int, int)[] = ( "+
				"        (1, 2), (1, 3), (1, 4), (1, 5), (1, 6), (1, 7), "+
				"        (2, 3), (2, 4), (2, 5), (2, 6), (2, 7), "+
				"        (3, 4), (3, 5), (3, 6), (3, 7), "+
				"        (4, 5), (4, 6), (4, 7), "+
				"        (5, 6), (5, 7), "+
				"        (6, 7) "+
				"    ); "+
				"    n:int = 7; "+
				"    lower_bound:int = 1; "+
				"    upper_bound:int = 100; "+
				"    k:int = lower_bound; "+
				"    number_of_solns:int = print_coloring(edges, start, n, k); "+
				"    while(number_of_solns == 0 & k < upper_bound){     "+
				"        k = k+1;     "+
				"        number_of_solns = print_coloring(edges, start, n, k); "+
				"    } "+
				"    print(\"# of colors=\");println(unparseInt(k)); "+
				"    print(\"# of different colorings=\");println(unparseInt(number_of_solns)); "+
				"}");
		try {
			if (getCygwinPath()==null)
				assertEquals("# of colors=7\n# of different colorings=5040\n", runBinaryWithInput(binary, ""));
			else
				assertEquals("# of colors=7\r\n# of different colorings=5040\r\n", runBinaryWithInput(binary, ""));
		} finally {
			binary.delete();
		}
	}
	
	@Test
	public void testGlobalTuple() {
		String output = compileLinkAndRun(
				"use io;\n" +
				"x:(int,int,bool);" +
				"main(args:int[][]) = {" +
				"	x=('A','B',true);\n" +
				"	(a:int,_,_)=x;" +
				"   println(\"Hello world!\");" +
				"	println(a);\n" +
				"}"
		);
		if (getCygwinPath()==null)
			assertEquals("Hello world!\nA\n", output);
		else
			assertEquals("Hello world!\r\nA\r\n", output);
	}
	
	@Test
	public void testShortCircuit() {
		String output =  compileLinkAndRun(
				"use io;" +
				"use conv;" +
				"use assert;" +
				"" +
				"unreachable():bool = {" +
				"    println(\"Unreachable function called!\");" +
				"    assert(false);" +
				"    return false;" +
				"}" +
				"" +
				"constFalse(): bool = {" +
				"    return false;" +
				"}" +
				"" +
				"constTrue(): bool = {" +
				"    return true;" +
				"}" +
				"" +
				"test(msg: int[], result: bool) = {" +
				"    print(\"Testing:\");" +
				"    println(msg);" +
				"    assert(result);" +
				"}" +
				"" +
				"main(args: int[][]) = {" +
				"    t: bool = constTrue();" +
				"    f: bool = constFalse();" +
				"    test(\"#1\", t | unreachable());" +
				"    test(\"#2\", !(f & unreachable()));" +
				"}");
		if (getCygwinPath()==null)
			assertEquals("Testing:#1\nTesting:#2\n", output);
		else
			assertEquals("Testing:#1\r\nTesting:#2\r\n", output);
	}
	
	@Test
	public void testGlobalArray() {
		String output = compileLinkAndRun(
				"use io;\n" +
				"x:int[];" +
				"main(args:int[][]) = {" +
				"	x=('A','B','C');\n" +
				"	print(x 0);\n" +
				"}"
		);
		assertEquals("A", output);
	}
	
	@Test
	public void testTupleWrite() {
		String output = compileLinkAndRun(
				"use io;\n" +
				"main(args:int[][]) = {" +
				"	x:(int,int,int)=('A','B','C');\n" +
				"	x 0 = 'D';" +
				"	println(x 0);\n" +
				"	a();" +
				"}"+
				"a() = {" +
				"	x:(int,int,int)=('E','F','G');\n" +
				"	x 0 = 'H';" +
				"	println(x 0);\n" +
				"}"
		);
		if (getCygwinPath()==null)
			assertEquals("D\nH\n", output);
		else
			assertEquals("D\r\nH\r\n", output);
	}
	
	@Test
	public void testGlobalTupleWrite() {
		String output = compileLinkAndRun(
				"use io;\n" +
				"x:(int,int,int);" +
				"main(args:int[][]) = {" +
				"	x=('A','B','C');\n" +
				"	x 0 = 'D';" +
				"	println(x 0);\n" +
				"	a();" +
				"}"+
				"a() = {" +
				"	x 0 = 'H';" +
				"	println(x 0);\n" +
				"}"
		);
		if (getCygwinPath()==null)
			assertEquals("D\nH\n", output);
		else
			assertEquals("D\r\nH\r\n", output);
	}

	@Test
	public void testSimpleClass() {
		String source =
		"use io;" +
		"use conv;" +
		"class Point {" +
		"	x, y: int;" +
		"   getX() : int = {" +
		"      return x;" +
		"   }" +
		"   getY() : int = {" +
		"      return this.y;" +
		"   }" +
		"   addXY() : int = {" +
		"      return (getX()+this.getY());" +
		"   }" +
		"}" +
		"createPoint(x:int, y:int): Point = {"+
		"	return initPoint(new Point, x, y);"+
		"}"+
		"initPoint(p:Point, x0:int, y0:int): Point = {"+
		"	 p.x = x0;"+
		"	 p.y = y0;"+
		"	 return p;" +
		"}"+
		"main(args:int[][]) = {" +
		"	cp:Point = createPoint(97,98);" +
		"	println(unparseInt(cp.getX()));" +
		"	println(unparseInt(cp.getY()));" +
		"	println(unparseInt(cp.addXY()));" +
		"}";
		String output = compileLinkAndRun(source);
		if (getCygwinPath()==null)
			assertEquals("97\n98\n195\n", output);
		else
			assertEquals("97\r\n98\r\n195\r\n", output);
	}

	@Test
	public void testInheritedClass() {
		String source =
		"use io; use assert;" +
		"class Point {" +
		"	x, y: int;" +
		"   getX() : int = {" +
		"      return x;" +
		"   }" +
		"   getY() : int = {" +
		"      return y;" +
		"   }" +
		"   addXY() : int = {" +
		"      return (this.getX()+this.getY());" +
		"   }" +
		"}" +
		"createPoint(x:int, y:int): Point = {"+
		"	return initPoint(new Point, x, y);"+
		"}"+
		"initPoint(p:Point, x0:int, y0:int): Point = {"+
		"	 p.x = x0;"+
		"	 p.y = y0;"+
		"	 return p;" +
		"}"+
		"class ColoredPoint extends Point {" +
		"	r,g,b:int;" +
		"}" +
		"createColoredPoint(x:int, y:int, r:int, g:int, b:int): ColoredPoint = {"+
		"	cp:ColoredPoint = new ColoredPoint;" +
		"	_=initPoint(cp, x,y); " +
		"	return initColoredPoint(cp,r,g,b); " +
		"}"+
		"initColoredPoint(p:ColoredPoint, r0:int, g0:int, b0:int): ColoredPoint = {"+
		"	 p.r = r0;"+
		"	 p.g = g0;"+
		"	 p.b = b0;"+
		"	 return p;" +
		"}"+
		"main(args:int[][]) = {" +
		"	cp:ColoredPoint = createColoredPoint(97,98,255,0,255);" +
		"	print(cp.getX());" +
		"	print(cp.getY());" +
		"	print(cp.addXY()-96);" +
		"   assert((cp.r+cp.g)==255);"+
		"}";
		String output = compileLinkAndRun(source);
		assertEquals("abc", output);
	}

	@Test
	public void testNestedTupleWrite() {
		String output = compileLinkAndRun(
				"use io;\n" +
				"main(args:int[][]) = {" +
				"	x:((int,int),(int,int))=(('A','B'),('C','D'));\n" +
				"	x 0 = ('E', 'F');" +
				"	print((x 0) 0);\n" +
				"}"
		);
		assertEquals("E", output);
	}

	@Test
	public void globalVariableTest() {
		String output = compileLinkAndRun(
				"use io;use conv;\n" +
				"A:int[2][i];" +
				"i:int = 3;" +
				"main(args:int[][]) = {" +
				"  A 1 2 = 3;" +
				"  println(unparseInt(i));" +
				"  i=42;" +
				"  println(unparseInt(i));" +
				"  println(unparseInt(length A));" +
				"  println(unparseInt(length (A 0)));" +
				"}"
		);
		if (getCygwinPath()==null)
			assertEquals("3\n42\n2\n3\n", output);
		else
			assertEquals("3\r\n42\r\n2\r\n3\r\n", output);
	}

	@Test
	public void testTupleAssignmentInArray() {
		String output = compileLinkAndRun(
			"use io;\n"+
			"use conv;\n"+
			"main(args:int[][]) = {"+
			"   a:(int,int)[2];"+
			"   a(0) = (1, 2);"+
			"   a(1) = (3, 4);"+
			"   a 0 0 = 5;"+
			"    print(unparseInt(a 0 0));"+
			"    print(unparseInt(a 0 1));"+
			"    print(unparseInt(a 1 0));"+
			"    print(unparseInt(a 1 1));"+
			"}");
		assertEquals("5234", output);
	}

	@Test
	public void testMestedTupleAssignment() {
		String output = compileLinkAndRun(
			"use io;\n"+
			"use conv;\n"+
			"main(args:int[][]) = {"+
			"   a:((int, int),(int, int));"+
			"   a(0) = (1, 2);"+
			"   a(1) = (3, 4);"+
			"   a 0 0 = 5;"+
			"    print(unparseInt(a 0 0));"+
			"    print(unparseInt(a 0 1));"+
			"    print(unparseInt(a 1 0));"+
			"    print(unparseInt(a 1 1));"+
			"}");
		assertEquals("5234", output);
	}
	/*
	@Test
	public void testLinkList() {
		String output = executeFilewithInput ("lists.i9", "5");
		if (getCygwinPath()==null)
			assertEquals("10000\n", output);
		else 
			assertEquals("10000\r\n", output);
	}
	
	@Test
	public void testRoots () {
		String output = executeFilewithInput("student-roots.i9","");
		assertEquals("", output);
	}
	
	@Test
	public void testAbstractFactoryPattern() {
		String output = executeFilewithInput("AbstractFactory.i9", "5");

		if (getCygwinPath() == null)
			assertEquals("Geeky ScrollBar has been created...\n" + "Size = 314",
					output);
		else
			assertEquals("Geeky ScrollBar has been created...\r\n" + "Size = 314",
					output);

	}
	*/
	public void testLexBench() {
		String input="";
		try {
			BufferedReader source = new BufferedReader(
					new FileReader("lex-bench.i9"));
			String line = source.readLine();
			while (line !=null) {
				input += line+"\n";
				line = source.readLine();
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
		String output = executeFilewithInput ("lex-bench.i9", input);
		try {
			BufferedReader refReader = new BufferedReader(
					new FileReader("lex-bench.ref"));
			BufferedReader outReader = new BufferedReader(
					new CharArrayReader(output.toCharArray()));
			String ref = refReader.readLine();
			String out = outReader.readLine();
			while (ref !=null && out != null) {
				assertEquals(ref, out);
				ref = refReader.readLine();
				out = outReader.readLine();
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
		try {
			FileWriter writer = new FileWriter ( "lex-bench.out");
			writer.append(output);
			writer.flush();
			writer.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
	}
	
	public void testObserverPattern() {
		String output = executeFilewithInput("observer.i9", "5");

		if (getCygwinPath() == null)
			assertEquals("Observer count = 1", output);
		else
			assertEquals("Observer count = 1", output);

	}

}
