package xic;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;
import static xic.parser.TokenType.AND;
import static xic.parser.TokenType.BOOL;
import static xic.parser.TokenType.BREAK;
import static xic.parser.TokenType.CHARACTER_LITERAL;
import static xic.parser.TokenType.CLOSE_BRACE;
import static xic.parser.TokenType.CLOSE_BRACKET;
import static xic.parser.TokenType.CLOSE_PAREN;
import static xic.parser.TokenType.COLON;
import static xic.parser.TokenType.COMMA;
import static xic.parser.TokenType.DIVIDE;
import static xic.parser.TokenType.ELSE;
import static xic.parser.TokenType.EOF;
import static xic.parser.TokenType.EQUAL;
import static xic.parser.TokenType.FALSE;
import static xic.parser.TokenType.GEQ;
import static xic.parser.TokenType.GETS;
import static xic.parser.TokenType.GT;
import static xic.parser.TokenType.IDENTIFIER;
import static xic.parser.TokenType.IF;
import static xic.parser.TokenType.INT;
import static xic.parser.TokenType.INTEGER_LITERAL;
import static xic.parser.TokenType.LENGTH;
import static xic.parser.TokenType.LEQ;
import static xic.parser.TokenType.LT;
import static xic.parser.TokenType.MINUS;
import static xic.parser.TokenType.MODULO;
import static xic.parser.TokenType.NOT;
import static xic.parser.TokenType.NOT_EQUAL;
import static xic.parser.TokenType.OPEN_BRACE;
import static xic.parser.TokenType.OPEN_BRACKET;
import static xic.parser.TokenType.OPEN_PAREN;
import static xic.parser.TokenType.OR;
import static xic.parser.TokenType.PLUS;
import static xic.parser.TokenType.RETURN;
import static xic.parser.TokenType.SEMICOLON;
import static xic.parser.TokenType.STRING_LITERAL;
import static xic.parser.TokenType.TIMES;
import static xic.parser.TokenType.TRUE;
import static xic.parser.TokenType.UNDERSCORE;
import static xic.parser.TokenType.USE;
import static xic.parser.TokenType.WHILE;

import java.io.StringReader;
import java.util.Iterator;

import org.junit.Before;
import org.junit.Test;

import xic.Configuration.Architecture;
import xic.lexer.BaseToken;
import xic.lexer.CharacterToken;
import xic.lexer.StringToken;
import xic.lexer.XiLexer;
import xic.parser.TokenType;
import edu.cornell.cs.cs4120.xi.CompilationException;

public class LexerTest {
	@Before
	public void initialize() {
		Configuration.setArchitecture(Architecture.I386);
	}

	/**
	 * Return a token iterator for a given string
	 */
	private Iterator<BaseToken> getTokenIterator(String content) {
		XiLexer lexer = new XiLexer("Test case", TokenType.SOURCE_IDENTIFIER, new StringReader(content));
		lexer.next(); // discard the first token
		return lexer;
	}

	/**
	 * Given an input string and an array of tokens, verify that the string
	 * creates this exact sequence of tokens
	 */
	public void runTest(String content, int[] tokens) {
		Iterator<BaseToken> lexer = getTokenIterator(content);
		int pos = 0;
		while (lexer.hasNext()) {
			BaseToken token = lexer.next();
			assertEquals("Token mismatch: got " + token.toString(), tokens[pos], token.type());
			pos++;
		}
		assertEquals("Mismatch - produced a different amount of tokens", pos,
				tokens.length);
	}
	

	@Test
	public void test1() {
		runTest("if (hello == \"hello\")", new int[] {
				IF, OPEN_PAREN,
				IDENTIFIER, EQUAL,
				STRING_LITERAL, CLOSE_PAREN });
	}

	@Test
	public void test2() {
		runTest("x:(bool, int) = (true, 1)", new int[] {
				IDENTIFIER, COLON,
				OPEN_PAREN, BOOL,
				COMMA, INT,
				CLOSE_PAREN, GETS,
				OPEN_PAREN, TRUE,
				COMMA, INTEGER_LITERAL,
				CLOSE_PAREN });
	}

	@Test
	public void test3() {
		runTest("s: int[] = \"Hello\";", new int[] {
				IDENTIFIER, COLON,
				INT, OPEN_BRACKET,
				CLOSE_BRACKET, GETS,
				STRING_LITERAL, SEMICOLON });
	}

	@Test
	public void test4() {
		runTest("gcd(a: int, b: int): int = {\n" 
				+ "  while (a != 0) {\n"
				+ "    if (a<b) b = b-a;\n"
				+ "    else a = a - b\n" + "  }\n"
				+ "  return b\n" + "}", new int[] {
				IDENTIFIER, OPEN_PAREN,
				IDENTIFIER, COLON,
				INT, COMMA,
				IDENTIFIER, COLON,
				INT, CLOSE_PAREN,
				COLON, INT, GETS, OPEN_BRACE, WHILE,
				OPEN_PAREN, IDENTIFIER,
				NOT_EQUAL, INTEGER_LITERAL,
				CLOSE_PAREN, OPEN_BRACE,
				IF, OPEN_PAREN,
				IDENTIFIER, LT,
				IDENTIFIER, CLOSE_PAREN,
				IDENTIFIER, GETS,
				IDENTIFIER, MINUS,
				IDENTIFIER, SEMICOLON,
				ELSE, IDENTIFIER,
				GETS, IDENTIFIER,
				MINUS, IDENTIFIER,
				CLOSE_BRACE, RETURN,
				IDENTIFIER, CLOSE_BRACE });
	}

	@Test
	public void test5() {
		String contentEscaped = "\"This\\n is \\\" a string with escapes \"";
		String content = "This\n is \" a string with escapes ";
		Iterator<BaseToken> lexer = getTokenIterator(contentEscaped);

		BaseToken token = lexer.next();
		assertEquals(token.type(), STRING_LITERAL);

		StringToken stringToken = (StringToken) token;

		assertEquals(stringToken.stringValue(), content);
		assertEquals(stringToken.value(), contentEscaped);
	}

	@Test
	public void test6() {
		runTest("a:int[] = (1,\n" 
				+ " 5, // this is a comment\n" 
				+ " 3);\n",
				new int[] { IDENTIFIER,
						COLON, INT,
						OPEN_BRACKET,
						CLOSE_BRACKET, GETS,
						OPEN_PAREN,
						INTEGER_LITERAL, COMMA,
						INTEGER_LITERAL, COMMA,
						INTEGER_LITERAL,
						CLOSE_PAREN, SEMICOLON });
	}

	@Test
	public void test7() {
		runTest("if ( a >= '\\'' ) return x else return y",
				new int[] { IF,
						OPEN_PAREN, IDENTIFIER,
						GEQ, CHARACTER_LITERAL,
						CLOSE_PAREN, RETURN,
						IDENTIFIER, ELSE,
						RETURN, IDENTIFIER });

		Iterator<BaseToken> lexer = getTokenIterator("'\\''");
		BaseToken token = lexer.next();
		assertEquals(token.type(), CHARACTER_LITERAL);
		assertEquals(token.value(), "'\\''");
		assertEquals(((CharacterToken) token).charValue(), '\'');
	}

	@Test
	public void test8() {
		Iterator<BaseToken> lexer = getTokenIterator(
				"\"this is a test of string position\"");
		BaseToken token = lexer.next();
		assertEquals(token.type(), STRING_LITERAL);
		assertEquals(token.position().lineStart(), 0);
		assertEquals(token.position().columnStart(), 0);
		assertEquals(token.position().lineEnd(), 0);
		assertEquals(token.position().columnEnd(), 34);
	}

	@Test(expected = CompilationException.class)
	public void test9() {
		/* Multi-line strings are not supported */
		Iterator<BaseToken> lexer = getTokenIterator("\"this is a test \nof multiline strings\"");
		lexer.next();
	}

	@Test
	public void test10() {
		runTest("(x:int, _, z:int) = (1,2,3)", new int[] {
				OPEN_PAREN, IDENTIFIER,
				COLON, INT, COMMA,
				UNDERSCORE, COMMA,
				IDENTIFIER, COLON,
				INT, CLOSE_PAREN,
				GETS, OPEN_PAREN,
				INTEGER_LITERAL, COMMA,
				INTEGER_LITERAL, COMMA,
				INTEGER_LITERAL, CLOSE_PAREN });
	}

	@Test
	public void test11() {
		Iterator<BaseToken> lexer = getTokenIterator("\"this is a backslash \\\\ \" ");
		BaseToken token = lexer.next();
		assertEquals(token.type(), STRING_LITERAL);
		assertEquals("this is a backslash \\ ", ((StringToken) token).stringValue());
	}

	@Test(expected = CompilationException.class)
	public void test12() {
		Iterator<BaseToken> lexer = getTokenIterator("\"this is an invalid escape \\   \"");
		lexer.next();
	}

	@Test(expected=CompilationException.class)
	public void test13() {
		Iterator<BaseToken> lexer = getTokenIterator("'\\e'");
		lexer.next();
	}

	@Test
	public void test14() {
		runTest("use library; " 
				+ "while ( a <= length b | (c+d) % 2 > 0) "
				+ "{ f = e*g/h; "
				+ "if (!more & a ) break; else more = false;}",
				new int[] { USE,
						IDENTIFIER, SEMICOLON,
						WHILE, OPEN_PAREN,
						IDENTIFIER, LEQ,
						LENGTH, IDENTIFIER,
						OR, OPEN_PAREN,
						IDENTIFIER, PLUS,
						IDENTIFIER, CLOSE_PAREN,
						MODULO, INTEGER_LITERAL,
						GT, INTEGER_LITERAL,
						CLOSE_PAREN, OPEN_BRACE,
						IDENTIFIER, GETS,
						IDENTIFIER, TIMES,
						IDENTIFIER, DIVIDE,
						IDENTIFIER, SEMICOLON,
						IF, OPEN_PAREN,
						NOT, IDENTIFIER,
						AND, IDENTIFIER,
						CLOSE_PAREN, BREAK,
						SEMICOLON, ELSE,
						IDENTIFIER, GETS,
						FALSE, SEMICOLON,
						CLOSE_BRACE });
	}

	public void test15() {
		/* Test that the EOF token is passed correctly */
		Iterator<BaseToken> lexer = getTokenIterator("1");
		assertEquals(lexer.next().type(), INTEGER_LITERAL);
		assertEquals(lexer.next().type(), EOF);
		assertEquals(lexer.next().type(), EOF);
	}

	@Test()
	public void test16() {
		/* Test the hasNext() interface */
		Iterator<BaseToken> lexer = getTokenIterator("1");
		assertTrue(lexer.hasNext());
		lexer.next();
		assertFalse(lexer.hasNext());
	}

	@Test(expected = CompilationException.class)
	public void test17() {
		/*
		 * Verify that an error is thrown when encountering an out-of-range
		 * integer
		 */
		Iterator<BaseToken> tokenIterator = getTokenIterator("2147483648");
		tokenIterator.next();
	}

	@Test(expected = CompilationException.class)
	public void test18() {
		/* Verify that an error is thrown when encountering an invalid character */
		Iterator<BaseToken> tokenIterator = getTokenIterator("$");
		tokenIterator.next();
	}

	@Test()
	public void test19() {
		/* Test the position implementation */
		Iterator<BaseToken> tokenIterator = getTokenIterator("\"12\\n3\" 123");
		BaseToken token = tokenIterator.next();
		assertEquals(token.position().lineStart(), 0);
		assertEquals(token.position().lineEnd(), 0);
		assertEquals(token.position().columnStart(), 0);
		assertEquals(token.position().columnEnd(), 6);

		token = tokenIterator.next();
		assertEquals(token.position().lineStart(), 0);
		assertEquals(token.position().lineEnd(), 0);
		assertEquals(token.position().columnStart(), 8);
		assertEquals(token.position().columnEnd(), 10);
	}
}

